import math

from fontTools.misc.transform import Identity
from PyQt5.QtCore import QLineF, QPointF, Qt
from PyQt5.QtGui import QBrush, QColor, QPainter, QPainterPath, QPen, QTransform

from defconQt.tools import platformSpecific
from defconQt.tools.drawing import (
    colorToQColor,
    drawTextAtPoint,
    ellipsePath,
    lozengePath,
    rectanglePath,
    trianglePath,
)

# ------
# Colors
# ------

_defaultColors = dict(
    # General
    # -------
    background=QColor(Qt.white),
    # Font
    # ----
    # guidelines
    fontGuideline=QColor.fromRgbF(1, 0, 0, 0.5),
    # Glyph
    # -----
    # contour fill
    glyphContourFill=QColor.fromRgbF(0.95, 0.95, 0.95, 0.3),
    # contour stroke
    glyphContourStroke=QColor(34, 34, 34),
    # component fill
    glyphComponentFill=QColor(90, 90, 90, 135),
    # component stroke
    glyphComponentStroke=QColor.fromRgbF(0, 0, 0, 1),
    # points
    glyphOnCurvePoints=QColor(4, 100, 166, 190),
    glyphOnCurveSmoothPoints=QColor(41, 172, 118, 190),
    glyphOffCurvePoints=QColor(116, 116, 116),
    glyphOtherPoints=QColor(140, 140, 140, 240),
    # anchors
    glyphAnchor=QColor(178, 102, 76, 200),
    # selection
    glyphSelection=QColor(145, 170, 196, 155),
    # guidelines
    glyphGuideline=QColor.fromRgbF(0.3, 0.4, 0.85, 0.5),
    # marker
    glyphBluesMarker=QColor(235, 191, 202, 225),
    # grid
    gridColor=QColor(220, 220, 220),
)


def defaultColor(name):
    return _defaultColors[name]


# ----------
# Primitives
# ----------


def drawLine(painter, x1, y1, x2, y2, lineWidth=0):
    painter.save()
    pen = painter.pen()
    if x1 == x2 or y1 == y2:
        painter.setRenderHint(QPainter.Antialiasing, False)
        # antialiased drawing blends a little in color with the background
        # reduce alpha before drawing aliased
        color = pen.color()
        color.setAlphaF(0.9 * color.alphaF())
        pen.setColor(color)
    pen.setWidthF(lineWidth)
    painter.setPen(pen)
    painter.drawLine(QPointF(x1, y1), QPointF(x2, y2))
    painter.restore()


def drawGlyphWithAliasedLines(painter, glyph):
    curvePath, lines = glyph.getRepresentation("TruFont.SplitLinesQPainterPath")
    painter.drawPath(curvePath)
    for x1, y1, x2, y2 in lines:
        drawLine(painter, x1, y1, x2, y2)


# ----
# Font
# ----

# Guidelines


def drawFontGuidelines(
    painter, glyph, scale, rect, drawLines=True, drawText=True, color=None
):
    """
    Draws the font guidelines of the Glyph_ *glyph* in the form of lines if
    *drawLines* is true and text if *drawText* is true using QPainter_
    *painter*.

    *rect* specifies the rectangle which the lines will be drawn in (usually,
    that of the glyph’s advance width).

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QPainter: http://doc.qt.io/qt-5/qpainter.html
    """
    if not (drawLines or drawText):
        return
    font = glyph.font
    if font is None:
        return
    if color is None:
        color = defaultColor("fontGuideline")
    _drawGuidelines(painter, glyph, scale, rect, font.guidelines, color=color)


def drawGlyphGuidelines(
    painter, glyph, scale, rect, drawLines=True, drawText=True, color=None
):
    if not (drawLines or drawText):
        return
    if color is None:
        color = defaultColor("glyphGuideline")
    _drawGuidelines(painter, glyph, scale, rect, glyph.guidelines, color=color)


def _drawGuidelines(
    painter,
    glyph,
    scale,
    rect,
    guidelines,
    drawLines=True,
    drawText=True,
    drawSelection=True,
    color=None,
):
    if not (drawLines or drawText):
        return
    xMin, yMin, width, height = rect
    xMax = xMin + width
    yMax = yMin + height
    for line in guidelines:
        color_ = color
        if color_ is None:
            if line.color:
                color_ = colorToQColor(line.color)
            else:
                color_ = defaultColor("glyphGuideline")
        painter.save()
        painter.setPen(color)
        line1 = None
        if None not in (line.x, line.y):
            if line.angle is not None:
                # make an infinite line that intersects *(line.x, line.y)*
                # 1. make horizontal line from *(line.x, line.y)* of length
                # *diagonal*
                diagonal = math.sqrt(width ** 2 + height ** 2)
                line1 = QLineF(line.x, line.y, line.x + diagonal, line.y)
                # 2. set the angle
                # defcon guidelines are clockwise
                line1.setAngle(line.angle)
                # 3. reverse the line and set length to 2 * *diagonal*
                line1.setPoints(line1.p2(), line1.p1())
                line1.setLength(2 * diagonal)
            else:
                line1 = QLineF(xMin, line.y, xMax, line.y)
        textX = 0
        textY = 0
        if drawLines:
            if line1 is not None:
                # line
                drawLine(painter, line1.x1(), line1.y1(), line1.x2(), line1.y2())
                # point
                x, y = line.x, line.y
                smoothWidth = 8 * scale
                smoothHalf = smoothWidth / 2.0
                painter.save()
                pointPath = QPainterPath()
                x -= smoothHalf
                y -= smoothHalf
                pointPath.addEllipse(x, y, smoothWidth, smoothWidth)
                pen = QPen(color_)
                pen.setWidthF(1 * scale)
                painter.setPen(pen)
                if drawSelection and line.selected:
                    painter.fillPath(pointPath, color_)
                painter.drawPath(pointPath)
                painter.restore()
            else:
                if line.y is not None:
                    drawLine(painter, xMin, line.y, xMax, line.y)
                elif line.x is not None:
                    drawLine(painter, line.x, yMin, line.x, yMax)
        if drawText and line.name:
            if line1 is not None:
                textX = line.x
                textY = line.y - 6 * scale
                xAlign = "center"
            else:
                if line.y is not None:
                    fontSize = painter.font().pointSize()
                    textX = glyph.width + 6 * scale
                    textY = line.y - (fontSize / 3.5) * scale
                elif line.x is not None:
                    textX = line.x + 6 * scale
                    textY = 0
                xAlign = "left"
            drawTextAtPoint(painter, line.name, textX, textY, scale, xAlign=xAlign)
        painter.restore()


# Image


def drawGlyphImage(painter, glyph, scale, drawSelection=True, selectionColor=None):
    image = glyph.image
    pixmap = image.getRepresentation("defconQt.QPixmap")
    if pixmap is None:
        return
    if selectionColor is None:
        selectionColor = defaultColor("glyphSelection")
    painter.save()
    painter.setTransform(QTransform(*image.transformation), True)
    painter.save()
    painter.translate(0, pixmap.height())
    painter.scale(1, -1)
    painter.drawPixmap(0, 0, pixmap)
    painter.restore()
    if drawSelection and image.selected:
        pen = QPen(selectionColor)
        pen.setWidthF(3.5 * scale)
        painter.setPen(pen)
        painter.drawRect(pixmap.rect())
    painter.restore()


# Fill and Stroke


def drawGlyphFillAndStroke(
    painter,
    glyph,
    scale,
    drawFill=True,
    drawStroke=True,
    drawSelection=True,
    drawComponentFill=True,
    drawComponentStroke=False,
    contourFillColor=None,
    contourStrokeColor=None,
    componentFillColor=None,
    componentStrokeColor=None,
    selectionColor=None,
    partialAliasing=True,
):
    if glyph.template:
        if glyph.unicode is None:
            return
        text = chr(glyph.unicode)
        font = glyph.font
        height = 750
        if font is not None and font.info.ascender:
            height = font.info.ascender
        painter.save()
        font = platformSpecific.otherUIFont()
        font.setPointSize(height)
        painter.setFont(font)
        color = QColor(Qt.lightGray)
        color.setAlphaF(0.4)
        painter.setPen(color)
        metrics = painter.fontMetrics()
        xOffset = -(metrics.width(text) - glyph.width) / 2
        painter.translate(xOffset, 0)
        painter.scale(1, -1)
        painter.drawText(0, 0, text)
        painter.restore()
        return
    # get the layer color
    layer = glyph.layer
    layerColor = None
    if layer is not None and layer.color is not None:
        layerColor = colorToQColor(layer.color)
    if selectionColor is None:
        selectionColor = defaultColor("glyphSelection")
    # get the paths
    contourPath = glyph.getRepresentation("defconQt.NoComponentsQPainterPath")
    componentPath, selectedComponentPath, originPts = glyph.getRepresentation(
        "TruFont.SelectedComponentsQPainterPath"
    )
    painter.save()
    # fill
    # contours
    if drawFill:
        if contourFillColor is None:
            if layerColor is not None:
                contourFillColor = layerColor
            else:
                contourFillColor = defaultColor("glyphContourFill")
        painter.fillPath(contourPath, QBrush(contourFillColor))
    # components
    if drawComponentFill:
        if componentFillColor is None:
            if layerColor is not None:
                componentFillColor = layerColor
            else:
                componentFillColor = defaultColor("glyphComponentFill")
        selectedComponentFillColor = QColor(componentFillColor)
        selectedComponentFillColor.setRed(0)
        selectedComponentFillColor.setGreen(0)
        selectedComponentFillColor.setBlue(0)
        painter.fillPath(componentPath, QBrush(componentFillColor))
        if drawSelection:
            painter.fillPath(selectedComponentPath, QBrush(selectedComponentFillColor))
        else:
            painter.fillPath(selectedComponentPath, QBrush(componentFillColor))
        # components origin
        # TODO: make this a parameter, disable on sizes < MinDetails
        if drawSelection:
            painter.save()
            pen = QPen(componentFillColor)
            pen.setWidth(0)
            painter.setPen(pen)
            for x, y in originPts:
                painter.drawLine(x, y + 5 * scale, x, y)
                painter.drawLine(x, y, x + 4.5 * scale, y)
            painter.restore()
    # selection
    if drawSelection:
        selectionPath = glyph.getRepresentation("TruFont.SelectedContoursQPainterPath")
        pen = QPen(selectionColor)
        pen.setWidthF(3.5 * scale)
        painter.setPen(pen)
        painter.drawPath(selectionPath)
    # stroke
    if drawStroke:
        # work out the color
        if contourStrokeColor is None:
            if layerColor is not None:
                contourStrokeColor = layerColor
            else:
                contourStrokeColor = defaultColor("glyphContourStroke")
        # contours
        pen = QPen(contourStrokeColor)
        pen.setWidth(0)
        painter.setPen(pen)
        if partialAliasing:
            drawGlyphWithAliasedLines(painter, glyph)
        else:
            painter.drawPath(contourPath)
    # components
    if drawComponentStroke:
        if componentStrokeColor is None:
            if layerColor is not None:
                componentStrokeColor = layerColor
            else:
                componentStrokeColor = defaultColor("glyphContourStroke")
        pen = QPen(componentStrokeColor)
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(selectedComponentPath)
        painter.drawPath(componentPath)
    painter.restore()


# points


def drawGlyphPoints(
    painter,
    glyph,
    scale,
    drawStartPoints=True,
    drawOnCurves=True,
    drawOffCurves=True,
    drawCoordinates=False,
    drawHandleCoordinates=False,
    drawCoordinatesOnSelection=False,
    drawSelection=True,
    drawBluesMarkers=True,
    onCurveColor=None,
    onCurveSmoothColor=None,
    offCurveColor=None,
    otherColor=None,
    backgroundColor=None,
):
    if onCurveColor is None:
        onCurveColor = defaultColor("glyphOnCurvePoints")
    if onCurveSmoothColor is None:
        onCurveSmoothColor = defaultColor("glyphOnCurveSmoothPoints")
    if offCurveColor is None:
        offCurveColor = defaultColor("glyphOffCurvePoints")
    if otherColor is None:
        otherColor = defaultColor("glyphOtherPoints")
    if backgroundColor is None:
        backgroundColor = defaultColor("background")
    bluesMarkerColor = defaultColor("glyphBluesMarker")
    notchColor = defaultColor("glyphContourStroke").lighter(200)
    # get the outline data
    outlineData = glyph.getRepresentation("defconQt.OutlineInformation")
    points = []
    handles = []
    # blue zones markers
    if drawBluesMarkers and drawOnCurves:
        font = glyph.font
        blues = []
        if font.info.postscriptBlueValues:
            blues += font.info.postscriptBlueValues
        if font.info.postscriptOtherBlues:
            blues += font.info.postscriptOtherBlues
        if blues:
            blues_ = set(blues)
            size = 13 * scale
            selectedSize = 15 * scale
            snapSize = 17 * scale
            selectedSnapSize = 20 * scale
            painter.save()
            pen = painter.pen()
            pen.setColor(QColor(255, 255, 255, 125))
            pen.setWidth(0)
            painter.setPen(pen)
            for point in outlineData["onCurvePoints"]:
                x, y = point["point"]
                # TODO: we could add a non-overlapping interval tree special
                # cased for borders
                selected = drawSelection and point.get("selected", False)
                if selected:
                    size_ = selectedSize
                    snapSize_ = selectedSnapSize
                else:
                    size_ = size
                    snapSize_ = snapSize
                for yMin, yMax in zip(blues[::2], blues[1::2]):
                    if not (y >= yMin and y <= yMax):
                        continue
                    # if yMin > 0 and y == yMin or yMin <= 0 and y == yMax:
                    if y in blues_:
                        path = lozengePath(x, y, snapSize_)
                    else:
                        path = ellipsePath(x, y, size_)
                    painter.fillPath(path, bluesMarkerColor)
                    painter.drawPath(path)
            painter.restore()
    # handles
    if drawOffCurves and outlineData["offCurvePoints"]:
        painter.save()
        painter.setPen(otherColor)
        for x1, y1, x2, y2 in outlineData["bezierHandles"]:
            handles.append((x2, y2))
            drawLine(painter, x1, y1, x2, y2)
        painter.restore()
    # on curve
    if drawOnCurves and outlineData["onCurvePoints"]:
        size = 6.5 * scale
        selectedSize = 8.5 * scale
        smoothSize = 8 * scale
        selectedSmoothSize = 10 * scale
        startSize = 7 * scale
        selectedStartSize = 9 * scale
        loneStartSize = 12 * scale
        selectedLoneStartSize = 14 * scale
        painter.save()
        notchPath = QPainterPath()
        paths = (QPainterPath(), QPainterPath())
        smoothPaths = (QPainterPath(), QPainterPath())
        for point in outlineData["onCurvePoints"]:
            x, y = point["point"]
            points.append((x, y))
            # notch
            if "smoothAngle" in point:
                angle = point["smoothAngle"]
                t = Identity.rotate(angle)
                x1, y1 = t.transformPoint((-1.35 * scale, 0))
                x2, y2 = -x1, -y1
                x1 += x
                y1 += y
                x2 += x
                y2 += y
                notchPath.moveTo(x1, y1)
                notchPath.lineTo(x2, y2)
            # points
            selected = drawSelection and point.get("selected", False)
            if selected:
                size_ = selectedSize
                smoothSize_ = selectedSmoothSize
                startSize_ = selectedStartSize
                loneStartSize_ = selectedLoneStartSize
            else:
                size_ = size
                smoothSize_ = smoothSize
                startSize_ = startSize
                loneStartSize_ = loneStartSize
            if drawStartPoints and "startPointAngle" in point:
                angle = point["startPointAngle"]
                if angle is not None:
                    pointPath = trianglePath(x, y, startSize_, angle)
                else:
                    pointPath = ellipsePath(x, y, loneStartSize_)
            elif point["smooth"]:
                pointPath = ellipsePath(x, y, smoothSize_)
            else:
                pointPath = rectanglePath(x, y, size_)
            # store the path
            if point["smooth"]:
                smoothPaths[selected].addPath(pointPath)
            else:
                paths[selected].addPath(pointPath)
            if drawCoordinatesOnSelection and selected and not drawCoordinates:
                drawPointText(painter, x, y, scale, ishandle=False)
        path, selectedPath = paths
        smoothPath, selectedSmoothPath = smoothPaths
        # fill
        selectedPath.setFillRule(Qt.WindingFill)
        selectedSmoothPath.setFillRule(Qt.WindingFill)
        painter.fillPath(selectedPath, onCurveColor)
        painter.fillPath(selectedSmoothPath, onCurveSmoothColor)
        # stroke
        pen = QPen(onCurveColor)
        pen.setWidthF(1.2 * scale)
        painter.setPen(pen)
        painter.drawPath(path)
        pen.setColor(onCurveSmoothColor)
        painter.setPen(pen)
        painter.drawPath(smoothPath)
        # notch
        pen.setColor(notchColor)
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(notchPath)
        painter.restore()
    # off curve
    if drawOffCurves and outlineData["offCurvePoints"]:
        # points
        offSize = 4.25 * scale
        selectedOffSize = 6.75 * scale
        path = QPainterPath()
        selectedPath = QPainterPath()
        selectedPath.setFillRule(Qt.WindingFill)
        for point in outlineData["offCurvePoints"]:
            x, y = point["point"]
            selected = drawSelection and point.get("selected", False)
            if selected:
                offSize_ = selectedOffSize
            else:
                offSize_ = offSize
            pointPath = ellipsePath(x, y, offSize_)
            if selected:
                selectedPath.addPath(pointPath)
            else:
                path.addPath(pointPath)
            if drawCoordinatesOnSelection and selected and not drawHandleCoordinates:
                drawPointText(painter, x, y, scale, ishandle=True)
        pen = QPen(offCurveColor)
        pen.setWidthF(2.5 * scale)
        painter.save()
        painter.setPen(pen)
        painter.drawPath(path)
        painter.fillPath(path, QBrush(backgroundColor))
        painter.fillPath(selectedPath, QBrush(offCurveColor.lighter(135)))
        painter.restore()
    # coordinates
    if drawCoordinates:
        painter.save()
        for x, y in points:
            drawPointText(painter, x, y, scale, ishandle=False)
        painter.restore()
    # handle coordinates
    if drawHandleCoordinates:
        painter.save()
        for x, y in handles:
            drawPointText(painter, x, y, scale, ishandle=True)
        painter.restore()


def drawPointText(painter, x, y, scale, ishandle=False):
    color = defaultColor("glyphOtherPoints")
    font = painter.font()
    font.setPointSize(7)
    if ishandle:
        color = color.lighter(135)
    painter.setPen(color)
    painter.setFont(font)

    posX = x
    # TODO: We use + here because we align on top. Consider abstracting
    # yOffset.
    posY = y + 6 * scale
    x = round(x, 1)
    if int(x) == x:
        x = int(x)
    y = round(y, 1)
    if int(y) == y:
        y = int(y)
    text = "%d  %d" % (x, y)

    drawTextAtPoint(painter, text, posX, posY, scale, xAlign="center", yAlign="top")


# Anchors


def drawGlyphAnchors(
    painter,
    glyph,
    scale,
    drawAnchors=True,
    drawSelection=True,
    drawText=True,
    color=None,
):
    if not glyph.anchors:
        return
    if color is None:
        color = defaultColor("glyphAnchor")
    fallbackColor = color
    anchorSize = 9 * scale
    selectedAnchorSize = 11 * scale
    for anchor in glyph.anchors:
        if anchor.color is not None:
            color = colorToQColor(anchor.color)
        else:
            color = fallbackColor
        x, y = anchor.x, anchor.y
        name = anchor.name
        painter.save()
        if drawAnchors:
            if drawSelection and anchor.selected:
                size = selectedAnchorSize
            else:
                size = anchorSize
            path = lozengePath(x, y, size)
            painter.fillPath(path, color)
        if drawText and name and drawSelection and anchor.selected:
            painter.setPen(color)
            # TODO: we're using + before we shift to top, ideally this should
            # be abstracted w drawTextAtPoint taking a dy parameter that will
            # offset the drawing region from origin regardless of whether we
            # are aligning to top or bottom.
            y += 6 * scale
            drawTextAtPoint(painter, name, x, y, scale, xAlign="center", yAlign="top")
        painter.restore()


# Grid


def drawGrid(painter, scale, rect, color=None):
    if color is None:
        color = defaultColor("gridColor")
    xMin, yMin, width, height = rect
    xMax = x = round(xMin + width)
    yMax = y = round(yMin + height)
    xMin, yMin = int(xMin), int(yMin)
    painter.save()
    pen = QPen(color)
    pen.setWidth(0)
    painter.setPen(pen)
    while x > xMin:
        painter.drawLine(x, yMin, x, yMax)
        x -= 1
    while y > yMin:
        painter.drawLine(xMin, y, xMax, y)
        y -= 1
    painter.restore()



from defcon import Component, Glyph, registerRepresentationFactory

from trufont.representationFactories.glyphCellFactory import TFGlyphCellFactory
from trufont.representationFactories.glyphViewFactory import (
    ComponentQPainterPathFactory,
    FilterSelectionFactory,
    SelectedComponentsQPainterPathFactory,
    SelectedContoursQPainterPathFactory,
    SplitLinesQPainterPathFactory,
)

# TODO: fine-tune the destructive notifications
_glyphFactories = {
    "TruFont.SelectedComponentsQPainterPath": (
        SelectedComponentsQPainterPathFactory,
        ("Glyph.Changed", "Glyph.SelectionChanged"),
    ),
    "TruFont.SplitLinesQPainterPath": (SplitLinesQPainterPathFactory, None),
    "TruFont.FilterSelection": (
        FilterSelectionFactory,
        ("Glyph.Changed", "Glyph.SelectionChanged"),
    ),
    "TruFont.SelectedContoursQPainterPath": (
        SelectedContoursQPainterPathFactory,
        ("Glyph.Changed", "Glyph.SelectionChanged"),
    ),
    "TruFont.GlyphCell": (TFGlyphCellFactory, None),
}
_componentFactories = {
    "TruFont.QPainterPath": (
        ComponentQPainterPathFactory,
        ("Component.Changed", "Component.BaseGlyphDataChanged"),
    )
}


def registerAllFactories():
    for name, (factory, destructiveNotifications) in _glyphFactories.items():
        registerRepresentationFactory(
            Glyph, name, factory, destructiveNotifications=destructiveNotifications
        )
    for name, (factory, destructiveNotifications) in _componentFactories.items():
        registerRepresentationFactory(
            Component, name, factory, destructiveNotifications=destructiveNotifications
        )



from math import sqrt

from fontTools.misc import bezierTools
from fontTools.pens.basePen import decomposeQuadraticSegment

# TODO: curve distance


def distance(x1, y1, x2, y2):
    dx = x2 - x1
    dy = y2 - y1
    return sqrt(dx * dx + dy * dy)


def lineProjection(x1, y1, x2, y2, x, y, ditchOutOfSegment=True):
    """
    Returns minimum distance between line p1, p2 and point p.
    Adapted from Grumdrig, http://stackoverflow.com/a/1501725/2037879.

    If *ditchOutOfSegment* is set, this function will return None if point p
    cannot be projected on the segment, i.e. if there's no line perpendicular
    to p1 p2 that intersects both p and a point of p1 p2.
    This is useful for certain GUI usages. Set by default.
    """
    l2 = (x2 - x1) ** 2 + (y2 - y1) ** 2
    if l2 == 0:
        return (x1, y1, 0.0)
    aX = x - x1
    aY = y - y1
    bX = x2 - x1
    bY = y2 - y1
    t = (aX * bX + aY * bY) / l2
    if ditchOutOfSegment:
        if t < 0:
            return (x1, y1, t)
        elif t > 1:
            return (x2, y2, t)
    projX = x1 + t * bX
    projY = y1 + t * bY
    return (projX, projY, t)


def lineDistance(x1, y1, x2, y2, x, y):
    projX, projY, _ = lineProjection(x1, y1, x2, y2, x, y)
    return distance(x, y, projX, projY)


# intersections


def curveIntersections(x1, y1, x2, y2, p1, p2, p3, p4):
    """
    Computes intersection between a cubic spline and a line segment.
    Adapted from: https://www.particleincell.com/2013/cubic-line-intersection/

    Takes four defcon points describing curve and four scalars describing line
    parameters.
    """

    bx, by = x1 - x2, y2 - y1
    m = x1 * (y1 - y2) + y1 * (x2 - x1)
    a, b, c, d = bezierTools.calcCubicParameters(
        (p1.x, p1.y), (p2.x, p2.y), (p3.x, p3.y), (p4.x, p4.y)
    )

    pc0 = by * a[0] + bx * a[1]
    pc1 = by * b[0] + bx * b[1]
    pc2 = by * c[0] + bx * c[1]
    pc3 = by * d[0] + bx * d[1] + m
    r = bezierTools.solveCubic(pc0, pc1, pc2, pc3)

    sol = []
    for t in r:
        if t < 0 or t > 1:
            continue
        s0 = a[0] * t ** 3 + b[0] * t ** 2 + c[0] * t + d[0]
        s1 = a[1] * t ** 3 + b[1] * t ** 2 + c[1] * t + d[1]
        if (x2 - x1) != 0:
            s = (s0 - x1) / (x2 - x1)
        else:
            s = (s1 - y1) / (y2 - y1)
        if s < 0 or s > 1:
            continue
        sol.append((s0, s1, t))
    return sol


def qcurveIntersections(x1, y1, x2, y2, *pts):
    """
    Computes intersection between a cubic spline and a line segment.
    Adapted from: https://www.particleincell.com/2013/cubic-line-intersection/

    Takes four defcon points describing curve and four scalars describing line
    parameters.
    """

    sol = []

    # PACK for fontTools
    points = []
    for pt in pts:
        points.append((pt.x, pt.y))

    p1 = (pts[0].x, pts[0].y)
    for p2, p3 in decomposeQuadraticSegment(points[1:]):
        bx, by = (y1 - y2), (x2 - x1)
        m = x1 * y2 - x2 * y1
        a, b, c = bezierTools.calcQuadraticParameters(p1, p2, p3)

        # prepare for next turn
        p1 = p3

        pc0 = bx * a[0] - by * a[1]
        pc1 = (bx * b[0] + by * b[1]) / pc0
        pc2 = (bx * c[0] + by * c[1] + m) / pc0
        r = bezierTools.solveQuadratic(pc0, pc1, pc2)

        for t in r:
            if t < 0 or t > 1:
                continue
            s0 = a[0] * (1 - t) ** 2 + b[0] * 2 * t * (1 - t) + c[0] * t ** 2
            s1 = a[1] * (1 - t) ** 2 + b[1] * 2 * t * (1 - t) + c[1] * t ** 2
            if (x2 - x1) != 0:
                s = (s0 - x1) / (x2 - x1)
            else:
                s = (s1 - y1) / (y2 - y1)
            if s < 0 or s > 1:
                continue
            sol.append((s0, s1, t))
    return sol


def lineIntersection(x1, y1, x2, y2, x3, y3, x4, y4):
    """
    Computes intersection point of two lines if any, and the t value on p1p2
    (attn: if you want to use the t value order your segments accordingly!)
    Adapted from Andre LaMothe, "Tricks of the Windows Game Programming Gurus".
    G. Bach, http://stackoverflow.com/a/1968345

    Takes four scalars describing line and four scalars describing otherLine.
    """

    Bx_Ax = x2 - x1
    By_Ay = y2 - y1
    Dx_Cx = x4 - x3
    Dy_Cy = y4 - y3
    determinant = -Dx_Cx * By_Ay + Bx_Ax * Dy_Cy
    if abs(determinant) < 1e-20:
        return None
    s = (-By_Ay * (x1 - x3) + Bx_Ax * (y1 - y3)) / determinant
    t = (Dx_Cx * (y1 - y3) - Dy_Cy * (x1 - x3)) / determinant
    if s >= 0 and s <= 1 and t >= 0 and t <= 1:
        return (x1 + (t * Bx_Ax), y1 + (t * By_Ay), t)
    return None



"""
The *baseWindows* submodule
---------------------------

The *baseWindows* submodule provides base implementations of top-level windows.

Due to some implementations details of memory management between Python and Qt,
parentless windows normally end up being deleted by the `Garbage Collector`_ of
Python and disappear, unless a reference to them is stored in a main namespace.

The base windows presented here take care of this problem automatically, and
should be used for spawning top-level windows (in scripting, notably).

- :class:`BaseWindow` is a QWidget_ top-level window
- :class:`BaseMainWindow` a QMainWindow_ top-level window

Both are used the same as their Qt widget parents, e.g.:

>>> from defconQt.windows.baseWindows import BaseWindow
>>> from PyQt5.QtWidgets import QLabel
>>> window = BaseWindow()
>>> label = QLabel("Hello World!", window)
>>> window.show()

Note: a QMainWindow_ is a QWidget subclass that has a predefined layout. See
its documentation for more details.

.. _QWidget: http://doc.qt.io/qt-5/qwidget.html
.. _QMainWindow: http://doc.qt.io/qt-5/qmainwindow.html
.. _`Garbage Collector`: https://en.wikipedia.org/wiki/Garbage_collection_(computer_science)
"""

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget


def _bootstrapGCCache(self):
    app = QApplication.instance()
    if not hasattr(app, "_windowCache"):
        app._windowCache = set()
    app._windowCache.add(self)


def _flushGCCache(self):
    app = QApplication.instance()
    try:
        app._windowCache.remove(self)
    except KeyError:
        # if self.close() is called more than once we'll end up
        # here. that's fine
        pass


class BaseMainWindow(QMainWindow):
    """
    A QMainWindow top-level window that keeps itself alive until it’s closed.

    See QMainWindow_ documentation for the list of available methods.

    .. _QMainWindow: http://doc.qt.io/qt-5/qmainwindow.html
    """

    def __init__(self, parent=None, flags=Qt.Window):
        super().__init__(parent, flags)
        self.setAttribute(Qt.WA_DeleteOnClose)
        _bootstrapGCCache(self)

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            _flushGCCache(self)


class BaseWindow(QWidget):
    """
    A QWidget top-level window that keeps itself alive until it’s closed.

    Sets Qt.Window as default *flags* parameter.

    See QWidget_ documentation for the list of available methods.

    .. _QWidget: http://doc.qt.io/qt-5/qwidget.html
    """

    def __init__(self, parent=None, flags=Qt.Window):
        super().__init__(parent, flags)
        self.setAttribute(Qt.WA_DeleteOnClose)
        _bootstrapGCCache(self)

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            _flushGCCache(self)



import functools
import itertools

import booleanOperations
from PyQt5.QtCore import QEvent, QLocale, QRegularExpression, QSize, Qt
from PyQt5.QtGui import QColor, QPainter, QRegularExpressionValidator
from PyQt5.QtWidgets import (
    QApplication,
    QDoubleSpinBox,
    QGridLayout,
    QHeaderView,
    QLineEdit,
    QScrollArea,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from defconQt.controls.colorVignette import ColorVignette
from defconQt.controls.listView import ListView
from defconQt.tools.drawing import colorToQColor
from trufont.controls.glyphAlignmentWidget import GlyphAlignmentWidget
from trufont.controls.groupBox import GroupBox
from trufont.controls.pathButton import PathButton
from trufont.objects import icons
from trufont.tools import platformSpecific
from trufont.tools.colorGenerator import ColorGenerator
from trufont.tools.rlabel import RLabel  # TODO: switch to QFormLayout


def Button(parent=None):
    btn = PathButton(parent)
    btn.setIsDownColor(QColor())
    btn.setFocusPolicy(Qt.NoFocus)
    btn.setSize(QSize(26, 26))
    return btn


def _glyphHasSelection(glyph):
    for contour in glyph:
        for point in contour:
            if point.selected:
                return True
    for container in (glyph.components, glyph.anchors, glyph.guidelines):
        for element in container:
            if element.selected:
                return True
    return False


def _partialTransform(glyph, matrix):
    glyph.beginUndoGroup()
    for contour in glyph:
        for point in contour:
            dirty = False
            if point.selected:
                point.x, point.y = matrix.transformPoint((point.x, point.y))
                dirty = True
            if dirty:
                contour.dirty = True
    for component in glyph.components:
        if component.selected:
            component.transform(matrix)
    for anchor in glyph.anchors:
        if anchor.selected:
            anchor.transform(matrix)
    for guideline in glyph.guidelines:
        if guideline.selected:
            guideline.transform(matrix)
    glyph.endUndoGroup()


class SpinBox(QDoubleSpinBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAlignment(Qt.AlignRight)
        self.setDecimals(3)

    def keyPressEvent(self, event):
        key_ = None
        if event.modifiers() & Qt.ShiftModifier:
            key = event.key()
            preDelta = None
            if key == Qt.Key_Up:
                key_ = Qt.Key_PageUp
                preDelta = 90
            elif key == Qt.Key_Down:
                key_ = Qt.Key_PageDown
                preDelta = -90
            if preDelta and event.modifiers() & Qt.ControlModifier:
                self.stepBy(preDelta)
        if key_ is not None:
            event = event.__class__(
                event.type(),
                key_,
                event.modifiers(),
                "",
                event.isAutoRepeat(),
                event.count(),
            )
        oldValue = self.value()
        super().keyPressEvent(event)
        if self.value() != oldValue:
            self.editingFinished.emit()

    def setValue(self, value):
        if value is None:
            self.clear()
            return
        super().setValue(value)

    def textFromValue(self, value):
        decimalPoint = QLocale().decimalPoint()
        if value.is_integer():
            return str(int(value))
        else:
            return str(value).replace(".", decimalPoint)


class NumberBox(SpinBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAlignment(Qt.AlignLeft)
        self.setButtonSymbols(QSpinBox.NoButtons)
        # TODO: MAX_INT?
        self.setRange(-900000, 900000)


class FillWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(240, 240, 240))


class PropertiesWidget(QWidget):
    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)

        self._font = font
        self._glyph = None
        self._shouldEditLastName = False

        glyphGroup = GroupBox(self)
        glyphGroup.setTitle(self.tr("Glyph"))
        glyphLayout = QGridLayout()
        zeroWidth = self.fontMetrics().width("0")
        columnOneWidth = zeroWidth * (5 + 2 * platformSpecific.widen())

        nameLabel = RLabel(self.tr("Name"), self)
        self.nameEdit = QLineEdit(self)
        self.nameEdit.editingFinished.connect(self.writeGlyphName)
        unicodesLabel = RLabel(self.tr("Unicode"), self)
        self.unicodesEdit = QLineEdit(self)
        self.unicodesEdit.editingFinished.connect(self.writeUnicodes)
        unicodesRegExp = QRegularExpression(
            "(|([a-fA-F0-9]{4,6})( ([a-fA-F0-9]{4,6}))*)"
        )
        unicodesValidator = QRegularExpressionValidator(unicodesRegExp, self)
        self.unicodesEdit.setValidator(unicodesValidator)
        widthLabel = RLabel(self.tr("Width"), self)
        self.widthEdit = NumberBox(self)
        self.widthEdit.setMaximumWidth(columnOneWidth)
        self.widthEdit.editingFinished.connect(self.writeWidth)
        leftMarginLabel = RLabel(self.tr("Left"), self)
        self.leftMarginEdit = NumberBox(self)
        self.leftMarginEdit.setMaximumWidth(columnOneWidth)
        self.leftMarginEdit.editingFinished.connect(self.writeLeftMargin)
        rightMarginLabel = RLabel(self.tr("Right"), self)
        self.rightMarginEdit = NumberBox(self)
        self.rightMarginEdit.setMaximumWidth(columnOneWidth)
        self.rightMarginEdit.editingFinished.connect(self.writeRightMargin)
        markColorLabel = RLabel(self.tr("Flag"), self)
        self.markColorWidget = ColorVignette(self)
        self.markColorWidget.colorChanged.connect(self.writeMarkColor)
        self.markColorWidget.setMaximumWidth(columnOneWidth)

        line = 0
        glyphLayout.addWidget(nameLabel, line, 0)
        glyphLayout.addWidget(self.nameEdit, line, 1, 1, 3)
        line += 1
        glyphLayout.addWidget(unicodesLabel, line, 0)
        glyphLayout.addWidget(self.unicodesEdit, line, 1, 1, 3)
        line += 1
        glyphLayout.addWidget(widthLabel, line, 0)
        glyphLayout.addWidget(self.widthEdit, line, 1)
        line += 1
        glyphLayout.addWidget(leftMarginLabel, line, 0)
        glyphLayout.addWidget(self.leftMarginEdit, line, 1)
        glyphLayout.addWidget(rightMarginLabel, line, 2)
        glyphLayout.addWidget(self.rightMarginEdit, line, 3)
        line += 1
        glyphLayout.addWidget(markColorLabel, line, 0)
        glyphLayout.addWidget(self.markColorWidget, line, 1)
        glyphLayout.setSpacing(8)
        glyphGroup.setChildLayout(glyphLayout)

        transformGroup = GroupBox(self)
        transformGroup.setTitle(self.tr("Transform"))
        transformLayout = QGridLayout()
        columnTwoWidth = zeroWidth * (8 + 2 * platformSpecific.widen())

        self.alignmentWidget = GlyphAlignmentWidget(self)
        self.alignmentWidget.setAlignment(4)

        # TODO: should this be implemented for partial selection?
        invScaleButton = Button(self)
        invScaleButton.setDrawingCommands(icons.dc_invscale())
        invScaleButton.setToolTip(self.tr("Scale down selection"))
        invScaleButton.clicked.connect(self.scale)
        invScaleButton.setProperty("inverted", True)
        self.scaleXEdit = SpinBox(self)
        self.scaleXEdit.setMaximumWidth(columnTwoWidth)
        self.scaleXEdit.setSuffix("%")
        self.scaleXEdit.setMaximum(500)
        self.scaleXEdit.setMinimum(-500)
        self.scaleXEdit.setValue(2)
        scaleButton = Button(self)
        scaleButton.setDrawingCommands(icons.dc_scale())
        scaleButton.setToolTip(self.tr("Scale up selection"))
        scaleButton.clicked.connect(self.scale)
        self.scaleYEdit = SpinBox(self)
        self.scaleYEdit.setMaximumWidth(columnTwoWidth)
        self.scaleYEdit.setSuffix("%")
        self.scaleYEdit.setMaximum(500)
        self.scaleYEdit.setMinimum(-500)
        self.scaleYEdit.setValue(2)

        rotateButton = Button(self)
        rotateButton.setDrawingCommands(icons.dc_rotate())
        rotateButton.setToolTip(self.tr("Rotate selection counter-clockwise"))
        rotateButton.clicked.connect(self.rotate)
        self.rotateEdit = SpinBox(self)
        self.rotateEdit.setMaximumWidth(columnTwoWidth)
        self.rotateEdit.setSuffix("º")
        # XXX: calling stepDown() from zero shows 359.999
        self.rotateEdit.setMaximum(359.999)
        self.rotateEdit.setValue(40)
        self.rotateEdit.setWrapping(True)
        invRotateButton = Button(self)
        invRotateButton.setDrawingCommands(icons.dc_invrotate())
        invRotateButton.setToolTip(self.tr("Rotate selection clockwise"))
        invRotateButton.clicked.connect(self.rotate)
        invRotateButton.setProperty("inverted", True)

        invSkewButton = Button(self)
        invSkewButton.setDrawingCommands(icons.dc_invskew())
        invSkewButton.setToolTip(self.tr("Skew selection counter-clockwise"))
        invSkewButton.clicked.connect(self.skew)
        invSkewButton.setProperty("inverted", True)
        self.skewEdit = SpinBox(self)
        self.skewEdit.setMaximumWidth(columnTwoWidth)
        self.skewEdit.setSuffix("º")
        self.skewEdit.setMaximum(100)
        self.skewEdit.setValue(6)
        self.skewEdit.setWrapping(True)
        skewButton = Button(self)
        skewButton.setDrawingCommands(icons.dc_skew())
        skewButton.setToolTip(self.tr("Skew selection clockwise"))
        skewButton.clicked.connect(self.skew)

        snapButton = Button(self)
        snapButton.setDrawingCommands(icons.dc_snap())
        snapButton.setToolTip(self.tr("Snap selection to precision"))
        snapButton.clicked.connect(self.snap)
        self.snapEdit = SpinBox(self)
        self.snapEdit.setMaximumWidth(columnTwoWidth)
        self.snapEdit.setValue(1)

        unionButton = Button(self)
        unionButton.setDrawingCommands(icons.dc_union())
        unionButton.setToolTip(self.tr("Remove selection overlap"))
        unionButton.clicked.connect(self.removeOverlap)
        subtractButton = Button(self)
        subtractButton.setDrawingCommands(icons.dc_subtract())
        subtractButton.setToolTip(self.tr("Subtract selected or top contour"))
        subtractButton.clicked.connect(self.subtract)
        intersectButton = Button(self)
        intersectButton.setDrawingCommands(icons.dc_intersect())
        intersectButton.setToolTip(self.tr("Intersect selected or top contour"))
        intersectButton.clicked.connect(self.intersect)
        xorButton = Button(self)
        xorButton.setDrawingCommands(icons.dc_xor())
        xorButton.setToolTip(self.tr("Xor selected or top contour"))
        xorButton.clicked.connect(self.xor)
        hMirrorButton = Button()
        hMirrorButton.setDrawingCommands(icons.dc_hmirror())
        hMirrorButton.setToolTip(self.tr("Mirror selection horizontally"))
        hMirrorButton.clicked.connect(self.hMirror)
        vMirrorButton = Button()
        vMirrorButton.setDrawingCommands(icons.dc_vmirror())
        vMirrorButton.setToolTip(self.tr("Mirror selection vertically"))
        vMirrorButton.clicked.connect(self.vMirror)

        alignHLeftButton = Button(self)
        alignHLeftButton.setDrawingCommands(icons.dc_alignhleft())
        alignHLeftButton.setToolTip(self.tr("Push selection left"))
        alignHLeftButton.clicked.connect(self.alignHLeft)
        alignHCenterButton = Button(self)
        alignHCenterButton.setDrawingCommands(icons.dc_alignhcenter())
        alignHCenterButton.setToolTip(self.tr("Push selection to horizontal center"))
        alignHCenterButton.clicked.connect(self.alignHCenter)
        alignHRightButton = Button(self)
        alignHRightButton.setDrawingCommands(icons.dc_alignhright())
        alignHRightButton.setToolTip(self.tr("Push selection right"))
        alignHRightButton.clicked.connect(self.alignHRight)
        alignVTopButton = Button(self)
        alignVTopButton.setDrawingCommands(icons.dc_alignvtop())
        alignVTopButton.setToolTip(self.tr("Push selection top"))
        alignVTopButton.clicked.connect(self.alignVTop)
        alignVCenterButton = Button(self)
        alignVCenterButton.setDrawingCommands(icons.dc_alignvcenter())
        alignVCenterButton.setToolTip(self.tr("Push selection to vertical center"))
        alignVCenterButton.clicked.connect(self.alignVCenter)
        alignVBottomButton = Button(self)
        alignVBottomButton.setDrawingCommands(icons.dc_alignvbottom())
        alignVBottomButton.setToolTip(self.tr("Push selection bottom"))
        alignVBottomButton.clicked.connect(self.alignVBottom)

        buttonsLayout = QGridLayout()
        buttonsLayout.setSpacing(0)
        line = 0
        buttonsLayout.addWidget(unionButton, line, 0)
        buttonsLayout.addWidget(subtractButton, line, 1)
        buttonsLayout.addWidget(intersectButton, line, 2)
        buttonsLayout.addWidget(xorButton, line, 3)
        buttonsLayout.addWidget(hMirrorButton, line, 4)
        buttonsLayout.addWidget(vMirrorButton, line, 5)
        line += 1
        buttonsLayout.addWidget(alignHLeftButton, line, 0)
        buttonsLayout.addWidget(alignHCenterButton, line, 1)
        buttonsLayout.addWidget(alignHRightButton, line, 2)
        buttonsLayout.addWidget(alignVTopButton, line, 3)
        buttonsLayout.addWidget(alignVCenterButton, line, 4)
        buttonsLayout.addWidget(alignVBottomButton, line, 5)

        line = 0
        transformLayout.addWidget(self.alignmentWidget, line, 1)
        line += 1
        transformLayout.addWidget(invScaleButton, line, 0)
        transformLayout.addWidget(self.scaleXEdit, line, 1)
        transformLayout.addWidget(scaleButton, line, 2)
        line += 1
        transformLayout.addWidget(self.scaleYEdit, line, 1)
        line += 1
        transformLayout.addWidget(rotateButton, line, 0)
        transformLayout.addWidget(self.rotateEdit, line, 1)
        transformLayout.addWidget(invRotateButton, line, 2)
        line += 1
        transformLayout.addWidget(invSkewButton, line, 0)
        transformLayout.addWidget(self.skewEdit, line, 1)
        transformLayout.addWidget(skewButton, line, 2)
        line += 1
        transformLayout.addWidget(snapButton, line, 0)
        transformLayout.addWidget(self.snapEdit, line, 1)
        line += 1
        transformLayout.addLayout(buttonsLayout, line, 0, 1, 3)
        transformLayout.setSpacing(4)
        transformGroup.setChildLayout(transformLayout)

        layersGroup = GroupBox(self)
        layersGroup.setTitle(self.tr("Layers"))
        layersLayout = QGridLayout()

        self.layerSetView = ListView(self)
        # QListView has no proper sizeHint by default
        self.layerSetView.sizeHint = lambda: QSize(150, 150)
        self.layerSetView.setDragEnabled(True)
        # HACK: we need this to setup headers and signals
        self.layerSetView.setList([[None, None, None]])
        hdr = self.layerSetView.header()
        hdr.setMinimumSectionSize(20)
        hdr.setStretchLastSection(False)
        hdr.setSectionResizeMode(QHeaderView.Fixed)
        hdr.setSectionResizeMode(1, QHeaderView.Stretch)
        hdr.resizeSection(0, 20)
        hdr.resizeSection(2, 34)
        self.layerSetView.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.layerSetView.dataDropped.connect(self.writeLayerOrder)
        self.layerSetView.valueChanged.connect(self.writeLayerAttribute)

        layerAddButton = Button(self)
        layerAddButton.setDrawingCommands(icons.dc_plus())
        layerAddButton.setToolTip(self.tr("Add a layer"))
        layerAddButton.clicked.connect(self.addLayer)
        layerRemoveButton = Button(self)
        layerRemoveButton.setDrawingCommands(icons.dc_minus())
        layerRemoveButton.setToolTip(self.tr("Remove selected layer"))
        layerRemoveButton.clicked.connect(self.removeLayer)
        layerDownButton = Button(self)
        layerDownButton.setDrawingCommands(icons.dc_down())
        layerDownButton.setToolTip(self.tr("Lower selected layer"))
        layerDownButton.clicked.connect(lambda: self.layerOffset(1))
        layerUpButton = Button(self)
        layerUpButton.setDrawingCommands(icons.dc_up())
        layerUpButton.setToolTip(self.tr("Raise selected layer"))
        layerUpButton.clicked.connect(lambda: self.layerOffset(-1))

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        layersLayout.setSpacing(0)
        line = 0
        layersLayout.addWidget(self.layerSetView, line, 0, 1, 5)
        line += 1
        layersLayout.addWidget(layerAddButton, line, 0)
        layersLayout.addWidget(layerRemoveButton, line, 1)
        layersLayout.addWidget(spacer, line, 2)
        layersLayout.addWidget(layerDownButton, line, 3)
        layersLayout.addWidget(layerUpButton, line, 4)
        layersGroup.setChildLayout(layersLayout)

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(glyphGroup)
        mainLayout.addWidget(transformGroup)
        mainLayout.addWidget(layersGroup)
        mainLayout.addWidget(FillWidget())
        mainLayout.setContentsMargins(0, 0, 0, 0)
        mainLayout.setSpacing(2)
        self.setLayout(mainLayout)

        self._updateLayerAttributes()
        self._subscribeToFont(self._font)

        # re-export signal
        self.activeLayerModified = self.layerSetView.selectionChanged_

    # -------------
    # Notifications
    # -------------

    def _unsubscribeFromGlyph(self):
        glyph = self._glyph
        if glyph is not None:
            glyph.removeObserver(self, "Glyph.Changed")

    def _subscribeToGlyph(self, glyph):
        if glyph is not None:
            glyph.addObserver(self, "_updateGlyphAttributes", "Glyph.Changed")

    def _unsubscribeFromFont(self):
        font = self._font
        if font is not None:
            layerSet = font.layers
            if layerSet is not None:
                layerSet.removeObserver(self, "LayerSet.Changed")

    def _subscribeToFont(self, font):
        if font is not None:
            layerSet = font.layers
            if layerSet is not None:
                layerSet.addObserver(self, "_updateLayerAttributes", "LayerSet.Changed")

    def _updateGlyph(self, *_):
        app = QApplication.instance()
        if app.currentFontWindow() != self.window():
            return
        self._unsubscribeFromGlyph()
        self._glyph = app.currentGlyph()
        self._subscribeToGlyph(self._glyph)
        self.alignmentWidget.setGlyph(self._glyph)
        self._updateGlyphAttributes()

    def _updateGlyphAttributes(self, *_):
        name = None
        unicodes = None
        width = None
        leftMargin = None
        rightMargin = None
        markColor = None
        if self._glyph is not None:
            glyph = self._glyph
            name = glyph.name
            if glyph.leftMargin is not None:
                leftMargin = round(glyph.leftMargin)
            if glyph.rightMargin is not None:
                rightMargin = round(glyph.rightMargin)
            foreGlyph = glyph.font[glyph.name]
            unicodes = " ".join(
                "%06X" % u if u > 0xFFFF else "%04X" % u for u in foreGlyph.unicodes
            )
            width = round(foreGlyph.width)
            if foreGlyph.markColor is not None:
                markColor = QColor.fromRgbF(*tuple(foreGlyph.markColor))

        self.nameEdit.setText(name)
        self.unicodesEdit.setText(unicodes)
        self.widthEdit.setValue(width)
        self.leftMarginEdit.setValue(leftMargin)
        self.rightMarginEdit.setValue(rightMargin)
        self.markColorWidget.setColor(markColor)

        self._updateLayerAttributes()

    def _updateLayerAttributes(self, *_):
        font = self._font
        if font is None or font.layers is None:
            self.layerSetView.blockSignals(True)
            self.layerSetView.setList([[None, None, None]])
            self.layerSetView.blockSignals(False)
            return
        layers = []
        currentLayer = self._glyph.layer if self._glyph is not None else None
        currentRow = -1
        for index, layer in enumerate(font.layers):
            color = layer.color
            if color is not None:
                color = colorToQColor(color)
            else:
                color = QColor()
            # XXX: add layer.visible
            layers.append([True, layer.name, color])
            if layer == currentLayer:
                currentRow = index
        self.layerSetView.blockSignals(True)
        self.layerSetView.setList(layers)
        self.layerSetView.setCurrentItem(currentRow, 0)
        self.layerSetView.blockSignals(False)
        if self._shouldEditLastName:
            self.layerSetView.editItem(len(layers) - 1, 1)
            self._shouldEditLastName = False

    # ---------
    # Callbacks
    # ---------

    # glyph attributes

    def writeGlyphName(self):
        if self._glyph is None:
            return
        newName = self.nameEdit.text()
        self._glyph.rename(newName)

    def writeUnicodes(self):
        if self._glyph is None:
            return
        text = self.unicodesEdit.text()
        if text:
            value = [int(uni, 16) for uni in text.split(" ")]
        else:
            value = []
        self._glyph.unicodes = value

    def writeWidth(self):
        if self._glyph is None:
            return
        self._glyph.width = self.widthEdit.value()

    def writeLeftMargin(self):
        if self._glyph is None:
            return
        self._glyph.leftMargin = self.leftMarginEdit.value()

    def writeRightMargin(self):
        if self._glyph is None:
            return
        self._glyph.rightMargin = self.rightMarginEdit.value()

    def writeMarkColor(self):
        color = self.markColorWidget.color()
        if color is not None:
            color = color.getRgbF()
        self._glyph.markColor = color

    def writeLayerAttribute(self, index, previous, current):
        font = self._font
        if font is None or font.layers is None:
            return
        row, column = index.row(), index.column()
        layers = font.layers
        layer = layers[layers.layerOrder[row]]
        if column == 0:
            pass  # TODO
        elif column == 1:
            # name
            layer.name = current
        elif column == 2:
            # color
            if current.isValid():
                color = current.getRgbF()
            else:
                color = None
            layer.color = color
        else:
            raise ValueError("invalid column selected: %d." % column)

    def writeLayerOrder(self):
        font = self._font
        if font is None or font.layers is None:
            return
        data = self.layerSetView.list()
        layerSet = font.layers
        layerOrder = []
        for _, name, _ in data:
            layerOrder.append(name)
        # defcon has data validity assertion (constant len etc.)
        layerSet.layerOrder = layerOrder

    # transforms

    def lockScale(self, checked):
        self.scaleYEdit.setEnabled(not checked)

    def scale(self):
        glyph = self._glyph
        # TODO: consider disabling the buttons in that case?
        if glyph is None:
            return
        sX = self.scaleXEdit.value()
        if not self.scaleYEdit.isEnabled():
            sY = sX
        else:
            sY = self.scaleYEdit.value()
        if self.sender().property("inverted"):
            sX, sY = -sX, -sY
        sX = 1 + 0.01 * sX
        sY = 1 + 0.01 * sY
        # apply
        hasSelection = _glyphHasSelection(glyph)
        representation = "TruFont.FilterSelection" if hasSelection else None
        center = self.alignmentWidget.origin(representation=representation)
        bareFunc = glyph.transform
        if hasSelection:
            glyph.transform = functools.partial(_partialTransform, glyph)
        glyph.scale((sX, sY), center=center)
        glyph.transform = bareFunc

    def rotate(self):
        glyph = self._glyph
        if glyph is None:
            return
        value = self.rotateEdit.value()
        if self.sender().property("inverted"):
            value = -value
        # apply
        hasSelection = _glyphHasSelection(glyph)
        representation = "TruFont.FilterSelection" if hasSelection else None
        origin = self.alignmentWidget.origin(representation=representation)
        bareFunc = glyph.transform
        if hasSelection:
            glyph.transform = functools.partial(_partialTransform, glyph)
        glyph.rotate(value, offset=origin)
        glyph.transform = bareFunc

    def skew(self):
        glyph = self._glyph
        if glyph is None:
            return
        value = self.skewEdit.value()
        if self.sender().property("inverted"):
            value = -value
        # apply
        hasSelection = _glyphHasSelection(glyph)
        representation = "TruFont.FilterSelection" if hasSelection else None
        origin = self.alignmentWidget.origin(representation=representation)
        bareFunc = glyph.transform
        if hasSelection:
            glyph.transform = functools.partial(_partialTransform, glyph)
        glyph.skew((value, 0), offset=origin)
        glyph.transform = bareFunc

    def snap(self):
        glyph = self._glyph
        if glyph is None:
            return
        base = self.snapEdit.value()
        # apply
        # TODO: this doesn't go by selection
        glyph.snap(base)

    # boolean ops

    def removeOverlap(self):
        # TODO: disable button instead
        glyph = self._glyph
        if not glyph:
            return
        target, others = [], []
        useSelection = bool(glyph.selection)
        for contour in glyph:
            if contour.open:
                others.append(contour)
                continue
            if useSelection and not contour.selection:
                others.append(contour)
                continue
            target.append(contour)
        glyph.beginUndoGroup()
        glyph.clearContours()
        pointPen = glyph.getPointPen()
        booleanOperations.union(target, pointPen)
        for contour in others:
            contour.drawPoints(pointPen)
        glyph.endUndoGroup()

    def _binaryBooleanOperation(self, func):
        # TODO: disable button instead
        glyph = self._glyph
        if glyph is None or len(glyph) < 2:
            return
        target, open_ = None, []
        delIndex = None
        others = list(glyph)
        for index, contour in enumerate(glyph):
            if contour.open:
                open_.append(contour)
            elif target is None and contour.selection:
                target = contour
                delIndex = index
        if delIndex is not None:
            del others[index]
        else:
            target = glyph[-1]
            del others[-1]
        glyph.beginUndoGroup()
        glyph.clearContours()
        pointPen = glyph.getPointPen()
        func(others, [target], pointPen)
        for contour in open_:
            contour.drawPoints(pointPen)
        glyph.endUndoGroup()

    def subtract(self):
        self._binaryBooleanOperation(booleanOperations.difference)

    def intersect(self):
        self._binaryBooleanOperation(booleanOperations.intersection)

    def xor(self):
        self._binaryBooleanOperation(booleanOperations.xor)

    # fitting

    def _mirror(self, attr, origin):
        glyph = self._glyph
        if not glyph:
            return
        points = glyph.selection
        useSelection = bool(points)
        vMin = vMax = None
        for contour in glyph:
            for point in contour:
                if useSelection and not point.selected:
                    continue
                value = getattr(point, attr)
                if vMin is None:
                    vMin = vMax = value
                else:
                    vMin = min(vMin, value)
                    vMax = max(vMax, value)
                if not useSelection:
                    points.add(point)
        factor = getattr(self.alignmentWidget, origin)()
        base = (2 - factor) * vMin + factor * vMax
        for point in points:
            value = base - getattr(point, attr)
            setattr(point, attr, value)
        # XXX: hack, relevant contours should be set dirty instead
        glyph.postNotification("Glyph.ContoursChanged")
        glyph.dirty = True

    def hMirror(self):
        self._mirror("x", "hOrigin")

    def vMirror(self):
        self._mirror("y", "vOrigin")

    def alignHLeft(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        xMin_all = None
        for contour in glyph:
            sel = False
            xMin = None
            for pt in contour:
                if xMin is None or xMin > pt.x:
                    xMin = pt.x
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, xMin))
                if xMin_all is None or xMin_all > xMin:
                    xMin_all = xMin
        if not selectedContours:
            return
        glyph.holdNotifications()
        for contour, xMin in selectedContours:
            if xMin > xMin_all:
                delta = xMin_all - xMin
                contour.move((delta, 0))
        glyph.releaseHeldNotifications()

    def alignHCenter(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        xMin_all, xMax_all = None, None
        for contour in glyph:
            sel = False
            xMin, xMax = None, None
            for pt in contour:
                if xMin is None or xMin > pt.x:
                    xMin = pt.x
                if xMax is None or xMax < pt.x:
                    xMax = pt.x
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, xMin, xMax))
                if xMin_all is None or xMin_all > xMin:
                    xMin_all = xMin
                if xMax_all is None or xMax_all < xMax:
                    xMax_all = xMax
        if not selectedContours:
            return
        glyph.holdNotifications()
        xAvg_all = xMin_all + round(0.5 * (xMax_all - xMin_all))
        for contour, xMin, xMax in selectedContours:
            xAvg = xMin + round(0.5 * (xMax - xMin))
            if xAvg != xAvg_all:
                delta = xAvg_all - xAvg
                contour.move((delta, 0))
        glyph.releaseHeldNotifications()

    def alignHRight(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        xMax_all = None
        for contour in glyph:
            sel = False
            xMax = None
            for pt in contour:
                if xMax is None or xMax < pt.x:
                    xMax = pt.x
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, xMax))
                if xMax_all is None or xMax_all < xMax:
                    xMax_all = xMax
        if not selectedContours:
            return
        glyph.holdNotifications()
        for contour, xMax in selectedContours:
            if xMax < xMax_all:
                delta = xMax_all - xMax
                contour.move((delta, 0))
        glyph.releaseHeldNotifications()

    def alignVTop(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        yMax_all = None
        for contour in glyph:
            sel = False
            yMax = None
            for pt in contour:
                if yMax is None or yMax < pt.y:
                    yMax = pt.y
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, yMax))
                if yMax_all is None or yMax_all < yMax:
                    yMax_all = yMax
        if not selectedContours:
            return
        glyph.holdNotifications()
        for contour, yMax in selectedContours:
            if yMax < yMax_all:
                delta = yMax_all - yMax
                contour.move((0, delta))
        glyph.releaseHeldNotifications()

    def alignVCenter(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        yMin_all, yMax_all = None, None
        for contour in glyph:
            sel = False
            yMin, yMax = None, None
            for pt in contour:
                if yMin is None or yMin > pt.y:
                    yMin = pt.y
                if yMax is None or yMax < pt.y:
                    yMax = pt.y
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, yMin, yMax))
                if yMin_all is None or yMin_all > yMin:
                    yMin_all = yMin
                if yMax_all is None or yMax_all < yMax:
                    yMax_all = yMax
        if not selectedContours:
            return
        glyph.holdNotifications()
        yAvg_all = yMin_all + round(0.5 * (yMax_all - yMin_all))
        for contour, yMin, yMax in selectedContours:
            yAvg = yMin + round(0.5 * (yMax - yMin))
            if yAvg != yAvg_all:
                delta = yAvg_all - yAvg
                contour.move((0, delta))
        glyph.releaseHeldNotifications()

    def alignVBottom(self):
        glyph = self._glyph
        if glyph is None:
            return
        selectedContours = []
        yMin_all = None
        for contour in glyph:
            sel = False
            yMin = None
            for pt in contour:
                if yMin is None or yMin > pt.y:
                    yMin = pt.y
                if pt.selected:
                    sel = True
            if sel:
                selectedContours.append((contour, yMin))
                if yMin_all is None or yMin_all > yMin:
                    yMin_all = yMin
        if not selectedContours:
            return
        glyph.holdNotifications()
        for contour, yMin in selectedContours:
            if yMin > yMin_all:
                delta = yMin_all - yMin
                contour.move((0, delta))
        glyph.releaseHeldNotifications()

    # layer operations

    def addLayer(self):
        font = self._font
        if font is None or font.layers is None:
            return
        name = "New layer"  # XXX: mangle
        self._shouldEditLastName = True
        layer = font.layers.newLayer(name)
        layer.color = tuple(itertools.chain(LayerColorGenerator.getColor(), (1,)))

    def removeLayer(self):
        font = self._font
        if font is None or font.layers is None:
            return
        if len(font.layers) <= 1:
            return
        name = self.layerSetView.currentRow()[1]
        del font.layers[name]

    def layerOffset(self, offset):
        font = self._font
        if font is None or font.layers is None:
            return
        data = self.layerSetView.list()
        i = self.layerSetView.currentIndex().row()
        ni = (i + offset) % len(data)
        layerSet = font.layers
        layerOrder = layerSet.layerOrder
        layerOrder[i], layerOrder[ni] = layerOrder[ni], layerOrder[i]
        # defcon has data validity assertion (constant len etc.)
        layerSet.layerOrder = layerOrder

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        # TODO: shouldn't size policy do this for us?
        return self.minimumSizeHint()

    def showEvent(self, event):
        super().showEvent(event)
        app = QApplication.instance()
        self._updateGlyph()
        app.dispatcher.addObserver(self, "_updateGlyph", "currentGlyphChanged")

    def hideEvent(self, event):
        super().hideEvent(event)
        app = QApplication.instance()
        app.dispatcher.removeObserver(self, "currentGlyphChanged")

    def closeEvent(self, event):
        self._unsubscribeFromFont()
        self._unsubscribeFromGlyph()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(212, 212, 212))


class PropertiesView(QScrollArea):
    propertiesWidgetClass = PropertiesWidget

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.setFrameShape(QScrollArea.NoFrame)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)

        self._propertiesWidget = self.propertiesWidgetClass(font, self)
        self.setWidget(self._propertiesWidget)

        self.activeLayerModified = self._propertiesWidget.activeLayerModified

    def eventFilter(self, obj, event):
        # this works because QScrollArea.setWidget installs an eventFilter
        # on the widget
        if obj == self.widget() and event.type() == QEvent.Resize:
            self.setMinimumWidth(
                self.widget().minimumSizeHint().width()
            )  # + self.verticalScrollBar().width())
        return super().eventFilter(obj, event)


# ---------------
# Color generator
# ---------------


class LayerColorGenerator(ColorGenerator):
    # precomputed colors fancy/k-means
    colors = [
        (185, 225, 122),
        (158, 206, 228),
        (233, 174, 200),
        (227, 191, 206),
        (130, 223, 184),
    ]
    index = 0

    @classmethod
    def getColor(cls):
        if cls.index <= len(cls.colors):
            color = (clr / 255 for clr in cls.colors[cls.index])
        else:
            color = ColorGenerator.getColor()
        cls.index += 1
        return color

    @classmethod
    def revert(cls):
        if cls.index > 0:
            cls.index -= 1



from collections import OrderedDict

from PyQt5.QtCore import QLineF, QPointF, Qt
from PyQt5.QtGui import QPainterPath
from PyQt5.QtWidgets import QApplication

from trufont.drawingTools.baseTool import BaseTool
from trufont.tools import bezierMath, drawing

_path = QPainterPath()
_path.moveTo(5.11, 23.16)
_path.lineTo(5.72, 24.04)
_path.lineTo(17.2, 12.82)
_path.lineTo(14.28, 9.9)
_path.lineTo(10.89, 10.72)
_path.closeSubpath()
_path.moveTo(15.53, 9.19)
_path.lineTo(18.19, 11.85)
_path.lineTo(22.96, 7.19)
_path.lineTo(19.99, 4.23)
_path.closeSubpath()


class KnifeTool(BaseTool):
    icon = _path
    name = QApplication.translate("KnifeTool", "Knife")
    shortcut = "E"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._cachedIntersections = None
        self._knifeLine = None
        self._knifePts = None

    def _appendIntersection(self, contour, index, pt):
        x, y, t = pt
        self._knifePts.append((x, y))
        if (contour, index) in self._cachedIntersections:
            self._cachedIntersections[(contour, index)].append(t)
        else:
            self._cachedIntersections[(contour, index)] = [t]

    def _findIntersections(self):
        self._cachedIntersections = OrderedDict()
        self._knifePts = []
        line = self._knifeLine
        for contour in self._glyph:
            segments = contour.segments
            for index, seg in enumerate(segments):
                if seg[-1].segmentType == "move":
                    continue
                prev = segments[index - 1][-1]
                if len(seg) == 3:
                    if seg[-1].segmentType == "qcurve":
                        i = bezierMath.qcurveIntersections(
                            line.x1(), line.y1(), line.x2(), line.y2(), prev, *seg
                        )
                    else:
                        i = bezierMath.curveIntersections(
                            line.x1(),
                            line.y1(),
                            line.x2(),
                            line.y2(),
                            prev,
                            seg[0],
                            seg[1],
                            seg[2],
                        )
                    for pt in i:
                        self._appendIntersection(contour, index, pt)
                elif len(seg) == 1:
                    pt = bezierMath.lineIntersection(
                        prev.x,
                        prev.y,
                        seg[0].x,
                        seg[0].y,
                        line.x1(),
                        line.y1(),
                        line.x2(),
                        line.y2(),
                    )
                    if pt is not None:
                        self._appendIntersection(contour, index, pt)

    # events

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            pos = event.localPos()
            self._knifeLine = QLineF(pos, pos)
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton:
            pos = event.localPos()
            if self._knifeLine is None:
                self._knifeLine = QLineF(pos, pos)
                return
            line = self._knifeLine
            if event.modifiers() & Qt.ShiftModifier:
                pos = self.clampToOrigin(pos, line.p1())
            line.setP2(pos)
            self._findIntersections()
            self.parent().update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() != Qt.LeftButton:
            super().mouseReleaseEvent(event)
            return
        if self._knifeLine is not None:
            p1, p2 = self._knifeLine.p1(), self._knifeLine.p2()
        self._knifeLine = None
        self._knifePts = None
        self.parent().update()
        # no-move clicks
        if not self._cachedIntersections:
            return
        self._glyph.beginUndoGroup()
        cutContours = (
            not event.modifiers() & Qt.AltModifier
            and len(self._cachedIntersections) > 1
        )
        if cutContours:
            oldPts = {pt for contour in self._glyph for pt in contour}
            path = self._glyph.getRepresentation("defconQt.QPainterPath")
        # reverse so as to not invalidate our cached segment indexes
        for loc, ts in reversed(list(self._cachedIntersections.items())):
            contour, index = loc
            prev = 1
            # reverse so as to cut from higher to lower value and compensate
            for t in sorted(ts, reverse=True):
                contour.splitAndInsertPointAtSegmentAndT(index, t / prev)
                prev = t
        # TODO: optimize
        if cutContours:
            newPts = {
                pt for contour in self._glyph for pt in contour if pt.segmentType
            } - oldPts
            del oldPts

            distances = dict()
            for point in newPts:
                d = bezierMath.distance(p1.x(), p1.y(), point.x, point.y)
                distances[d] = point
            del newPts

            sortedPts = [distances[dist] for dist in sorted(distances.keys())]
            del distances

            # group points by belonging to contour "black area"
            siblings = []
            stack = None
            if not path.contains(p1):
                stack = []
            for pt, nextPt in zip(sortedPts, sortedPts[1:]):
                qPt = QPointF(pt.x, pt.y)
                qHalf = qPt + 0.5 * (QPointF(nextPt.x, nextPt.y) - qPt)
                if path.contains(qHalf):
                    if stack is not None:
                        stack.append(pt)
                else:
                    if stack:
                        stack.append(pt)
                        siblings.extend(stack)
                    stack = []
            if not path.contains(p2):
                if stack:
                    stack.append(nextPt)
                    siblings.extend(stack)
            del stack

            # ok, now i = siblings.index(loc); siblings[i+1-2(i%2)] will yield
            # sibling
            newGlyph = self._glyph.__class__()
            pen = newGlyph.getPointPen()

            def _visitPath(contour, index):
                didJump = False
                pen.beginPath()
                while True:
                    pt = contour.getPoint(index)
                    if pt in visited:
                        pen.endPath()
                        break
                    if pt not in siblings:
                        visited.add(pt)
                    segmentType = pt.segmentType
                    smooth = pt.smooth
                    if didJump or pt in siblings:
                        smooth = False
                        if didJump:
                            segmentType = "line"
                    pen.addPoint((pt.x, pt.y), segmentType, smooth)
                    if pt in siblings and not didJump:
                        i = siblings.index(pt)
                        otherPt = siblings[i + 1 - 2 * (i % 2)]
                        # TODO: optimize-out this lookup
                        for c in self._glyph:
                            try:
                                index = c.index(otherPt)
                            except Exception:
                                pass
                            else:
                                contour = c
                                break
                        didJump = True
                        continue
                    didJump = False
                    index += 1

            visited = set()
            for contour in self._glyph:
                for index, pt in enumerate(contour):
                    if pt in visited or pt in siblings:
                        continue
                    _visitPath(contour, index)
            self._glyph.clearContours()
            pen = self._glyph.getPointPen()
            newGlyph.drawPoints(pen)
        self._glyph.endUndoGroup()
        ##
        self._cachedIntersections = None

    # custom painting

    def paint(self, painter, index):
        widget = self.parent()
        if index != widget.activeIndex():
            return
        line = self._knifeLine
        if line is not None and line.length():
            painter.save()
            pen = painter.pen()
            pen.setWidth(0)
            painter.setPen(pen)
            drawing.drawLine(painter, line.x1(), line.y1(), line.x2(), line.y2())
            if self._knifePts is not None:
                scale = widget.inverseScale()
                dotSize = 5 * scale
                dotHalf = dotSize / 2
                path = QPainterPath()
                for x, y in self._knifePts:
                    x -= dotHalf
                    y -= dotHalf
                    path.addEllipse(x, y, dotSize, dotSize)
                painter.drawPath(path)
            painter.restore()



import sys

from PyQt5.QtCore import QObject, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QPalette, QTextBlockFormat, QTextCursor, QTextOption
from PyQt5.QtWidgets import QCheckBox, QMainWindow, QPlainTextEdit, QPushButton

from defconQt.tools import platformSpecific as basePlatformSpecific
from trufont.objects import settings
from trufont.tools import platformSpecific


class OutputWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent, Qt.Tool)
        self.outputEdit = OutputEdit(self)
        palette = self.outputEdit.palette()
        palette.setColor(QPalette.Base, Qt.black)
        palette.setColor(QPalette.Text, Qt.white)
        self.outputEdit.setPalette(palette)
        self.outputEdit.viewport().setCursor(Qt.ArrowCursor)
        self.wrapLinesBox = QCheckBox(self.tr("Wrap Lines"), self)
        self.wrapLinesBox.toggled.connect(self.setWordWrapEnabled)
        clearOutputButton = QPushButton(self.tr("Clear"), self)
        clearOutputButton.clicked.connect(self.outputEdit.clear)

        self.setCentralWidget(self.outputEdit)
        self.setWindowTitle(self.tr("Output"))
        statusBar = self.statusBar()
        statusBar.addWidget(self.wrapLinesBox)
        statusBar.addPermanentWidget(clearOutputButton)
        statusBar.setSizeGripEnabled(False)
        if platformSpecific.needsTighterMargins():
            margins = (7, -10, 9, -12)
        else:
            margins = (4, -1, 5, 0)
        statusBar.setContentsMargins(*margins)

        for channel in ("stdout", "stderr"):
            stream = OutputStream(channel, self)
            stream.messagePassed.connect(self.outputEdit.write)

        self.readSettings()

    def readSettings(self):
        geometry = settings.outputWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)
        checked = settings.outputWindowWrapLines()
        self.wrapLinesBox.setChecked(checked)

    def writeSettings(self):
        settings.setOutputWindowGeometry(self.saveGeometry())
        settings.setOutputWindowWrapLines(self.wrapLinesBox.isChecked())

    def setWordWrapEnabled(self, value):
        if value:
            wrapMode = QTextOption.WrapAtWordBoundaryOrAnywhere
        else:
            wrapMode = QTextOption.NoWrap
        self.outputEdit.setWordWrapMode(wrapMode)
        self.writeSettings()

    # ----------
    # Qt methods
    # ----------

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def sizeHint(self):
        return QSize(465, 400)


class OutputEdit(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFont(basePlatformSpecific.fixedFont())
        self.setReadOnly(True)
        self.setUndoRedoEnabled(False)

    def grayOutOldContent(self):
        insert = True
        textCursor = self.textCursor()
        if not textCursor.atEnd():
            textCursor.movePosition(QTextCursor.End)
        if textCursor.atStart():
            insert = False
        endFormat = textCursor.charFormat()

        textCursor.select(QTextCursor.Document)

        fmt = endFormat.__class__()
        bgColor = self.palette().base().color()
        fgColor = self.palette().text().color()
        bgFactor = 0.5
        fgFactor = 1 - bgFactor
        fmt.setForeground(
            QColor(
                bgFactor * bgColor.red() + fgFactor * fgColor.red(),
                bgFactor * bgColor.green() + fgFactor * fgColor.green(),
                bgFactor * bgColor.blue() + fgFactor * fgColor.blue(),
            )
        )
        textCursor.mergeCharFormat(fmt)

        textCursor.movePosition(QTextCursor.End)
        textCursor.setCharFormat(endFormat)
        if insert:
            textCursor.insertBlock(QTextBlockFormat())
        self.setTextCursor(textCursor)

    def isScrollBarAtBottom(self):
        scrollBar = self.verticalScrollBar()
        return scrollBar.value() == scrollBar.maximum()

    def scrollToBottom(self):
        scrollBar = self.verticalScrollBar()
        scrollBar.setValue(scrollBar.maximum())
        # QPlainTextEdit destroys the first calls value in case of multiline
        # text, so make sure that the scroll bar actually gets the value set.
        # Is a noop if the first call succeeded.
        scrollBar.setValue(scrollBar.maximum())

    def write(self, message, stream=None):
        atBottom = self.isScrollBarAtBottom()
        textCursor = self.textCursor()
        if not textCursor.atEnd():
            self.moveCursor(QTextCursor.End)
        charFormat = self.currentCharFormat()
        if stream == "stderr":
            color = Qt.red
        else:
            color = self.palette().color(QPalette.Text)
        charFormat.setForeground(color)
        self.setCurrentCharFormat(charFormat)
        self.insertPlainText(message)
        if atBottom:
            self.scrollToBottom()


class OutputStream(QObject):
    messagePassed = pyqtSignal(str, str)

    def __init__(self, stream=None, parent=None):
        super().__init__(parent)
        self.forward = False
        self.stream = None
        self.oldStream = None
        self.registerStream(stream)

    def registerStream(self, stream):
        self.stream = stream
        if self.stream is None:
            return
        assert self.stream in ("stdout", "stderr")
        self.oldStream = getattr(sys, self.stream)
        setattr(sys, self.stream, self)

    def unregisterStream(self):
        if self.oldStream is None:
            return
        setattr(sys, self.stream, self.oldStream)
        self.stream = None
        self.oldStream = None

    def flush(self):
        pass

    def write(self, message):
        self.messagePassed.emit(message, self.stream)
        if self.forward:
            self.oldStream.write(message)



from PyQt5.QtCore import QMimeData


class GlyphsMimeData(QMimeData):
    def __init__(self):
        super().__init__()
        # we don't need to serialize until we want to do cross-app dnd
        self._glyphs = []

    def glyphs(self):
        return self._glyphs

    def setGlyphs(self, glyphs):
        self._glyphs = glyphs

    # ----------
    # Qt methods
    # ----------

    def formats(self):
        formats = super().formats()
        formats.append("text/plain")
        return formats

    def hasFormat(self, format_):
        if format_ == "text/plain":
            return True
        return super().hasFormat(format_)

    def retrieveData(self, mimeType, type_):
        if mimeType == "text/plain":
            return " ".join(glyph.name for glyph in self._glyphs)
        return super().retrieveData(mimeType, type_)



from PyQt5.QtCore import QEvent, QPoint, QPointF, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QCursor, QPainter
from PyQt5.QtWidgets import QApplication, QWidget

from defconQt.controls.glyphView import GlyphViewMinSizeForDetails, UIFont
from defconQt.tools import drawing, platformSpecific

# TODO: forbid scrolling past scene boundary

_noPointSizePadding = 200


class GlyphFlags:
    __slots__ = ["_isActiveGlyph", "_isActiveLayer"]

    def __init__(self, isActiveGlyph, isActiveLayer=True):
        self._isActiveGlyph = isActiveGlyph
        self._isActiveLayer = isActiveLayer

    def __repr__(self):
        return "<{} isActiveGlyph: {} isActiveLayer: {}>".format(
            self.__class__.__name__, self.isActiveGlyph, self.isActiveLayer
        )

    @property
    def isActiveGlyph(self):
        return self._isActiveGlyph

    @property
    def isActiveLayer(self):
        return self._isActiveLayer


class GlyphContextView(QWidget):
    activeGlyphChanged = pyqtSignal()
    pointSizeModified = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.DefaultContextMenu)
        self.setFocusPolicy(Qt.ClickFocus)
        self.grabGesture(Qt.PanGesture)
        self.grabGesture(Qt.PinchGesture)
        self._drawingOffset = QPoint()
        self._fitViewport = True
        self._glyphRecords = []
        self._glyphRecordsRects = {}
        self._activeIndex = 0

        # drawing attributes
        self._defaultDrawingAttributes = dict(
            showGlyphFill=False,
            showGlyphStroke=True,
            showGlyphComponentFill=True,
            showGlyphComponentStroke=False,
            showGlyphOnCurvePoints=True,
            showGlyphStartPoints=True,
            showGlyphOffCurvePoints=True,
            showGlyphPointCoordinates=False,
            showGlyphBezierHandlesCoordinates=False,
            showGlyphCoordinatesWhenSelected=False,
            showGlyphAnchors=True,
            showGlyphMetrics=True,
            showGlyphGuidelines=True,
            showGlyphImage=True,
            showFontVerticalMetrics=True,
            showFontVerticalMetricsTitles=False,
            showFontGuidelines=True,
            showFontPostscriptBlues=True,
            showFontPostscriptFamilyBlues=False,  # TODO: test appearance of
            # this combined w/ blues
        )

        # cached vertical metrics
        self._unitsPerEm = 1000
        self._descender = -250
        self._capHeight = 750
        self._ascender = 750

        # drawing data cache
        self._scale = 1.0
        self._inverseScale = 0.1
        self._impliedPointSize = 1000

        self._backgroundColor = Qt.white

    @property
    def _glyph(self):
        if not self._glyphRecords:
            return None
        return self._glyphRecords[self._activeIndex].glyph

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyphs(self, glyphRecords):
        handledGlyphs = set()
        handledFonts = set()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.addObserver(self, "_fontChanged", "Info.Changed")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        handledFonts = set()
        glyphRecords = self.glyphRecords()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.removeObserver(self, "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.removeObserver(self, "Info.Changed")

    def _glyphChanged(self, notification):
        self.update()

    def _fontChanged(self, notification):
        self.setGlyphRecords(self.glyphRecords())

    # --------------
    # Custom Methods
    # --------------

    def activeGlyph(self):
        return self._glyph

    def setActiveGlyph(self, glyph):
        glyphs = list(self.glyphs())
        glyphs[self._activeIndex] = glyph
        self.setGlyphs(glyphs)

    def activeIndex(self):
        return self._activeIndex

    def setActiveIndex(self, value):
        if value == self._activeIndex:
            return
        self._activeIndex = value
        self.activeGlyphChanged.emit()
        self.update()

    def glyphForIndex(self, index):
        if not self._glyphRecords:
            return None
        return self._glyphRecords[index].glyph

    def indexForPoint(self, point):
        if not self._glyphRecordsRects:
            return None
        for rect, recordIndex in self._glyphRecordsRects.items():
            # we don't bound the height here
            x, _, w, _ = rect
            if point.x() >= x and point.x() <= x + w:
                return recordIndex
        return None

    def originForIndex(self, index=None):
        if index is None:
            index = self._activeIndex
        ret = QPointF(self._drawingOffset)
        for glyphIndex, glyphRecord in enumerate(self._glyphRecords):
            if glyphIndex == index:
                xO, yO = glyphRecord.xOffset, glyphRecord.yOffset
                ret += QPointF(xO * self._scale, yO * self._scale)
                break
            xA = glyphRecord.xAdvance
            yA = glyphRecord.yAdvance
            ret += QPointF(xA * self._scale, yA * self._scale)
        return ret

    def inverseScale(self):
        return self._inverseScale

    def pointSize(self):
        return self._impliedPointSize

    def setPointSize(self, pointSize):
        scale = pointSize / self._unitsPerEm
        self.setScale(scale)

    def scale(self):
        return self._scale

    def setScale(self, scale):
        self._scale = scale
        if self._scale <= 0:
            self._scale = 0.01
        self._inverseScale = 1.0 / self._scale
        self._impliedPointSize = self._unitsPerEm * self._scale
        self.update()

    def glyphRecords(self):
        return self._glyphRecords

    def setGlyphRecords(self, glyphRecords):
        self._unsubscribeFromGlyphs()
        self._glyphRecords = glyphRecords
        self._font = None
        # XXX: for now, we assume all glyphs come from
        # the same font
        if self._glyphRecords:
            font = self._font = self._glyphRecords[0].glyph.font
            if font is not None:
                self._unitsPerEm = font.info.unitsPerEm
                if self._unitsPerEm is None:
                    self._unitsPerEm = 1000
                self._descender = font.info.descender
                if self._descender is None:
                    self._descender = -250
                self._ascender = font.info.ascender
                if self._ascender is None:
                    self._ascender = self._unitsPerEm + self._descender
                self._capHeight = font.info.capHeight
                if self._capHeight is None:
                    self._capHeight = self._ascender
            self.setScale(self._scale)
            if self._activeIndex >= len(self._glyphRecords):
                self._activeIndex = len(self._glyphRecords) - 1
        else:
            self._activeIndex = 0
        self._subscribeToGlyphs(glyphRecords)
        self.activeGlyphChanged.emit()
        self.update()

    def glyphs(self):
        for glyphRecord in self._glyphRecords:
            yield glyphRecord.glyph

    def setGlyphs(self, glyphs):
        glyphRecords = []
        for index, glyph in enumerate(glyphs):
            glyphRecord = GlyphRecord()
            glyphRecord.glyph = glyph
            glyphRecord.cluster = index
            #
            w, h = glyph.width, glyph.height
            layerSet = glyph.layerSet
            if layerSet is not None:
                layer = layerSet.defaultLayer
                if glyph.name in layer:
                    layerGlyph = layer[glyph.name]
                    w, h = layerGlyph.width, layerGlyph.height
            glyphRecord.xAdvance = w
            glyphRecord.yAdvance = h
            #
            glyphRecords.append(glyphRecord)
        self.setGlyphRecords(glyphRecords)

    # fitting

    def fitScaleMetrics(self):
        """
        Scales and centers the viewport around the font’s metrics.
        """
        fitHeight = self.height()
        fitWidth = self.width()
        glyph = self._glyph
        bottom, top = self.verticalBounds()
        height = -bottom + top
        self.setScale(fitHeight / height)
        otherWidth = 0
        for glyph_ in self.glyphs():
            if glyph_ == glyph:
                break
            otherWidth += glyph_.width
        dx = 0.5 * (fitWidth - glyph.width * self._scale) - otherWidth * self._scale
        dy = 0.5 * (fitHeight - height * self._scale) + top * self._scale
        # TODO: round?
        self._drawingOffset = QPoint(dx, dy)
        self.pointSizeModified.emit(self._impliedPointSize)

    def fitScaleBBox(self):
        """
        Scales and centers the viewport around the *glyph*’s bounding box.
        """
        glyph = self.activeGlyph()
        if glyph is None:
            return
        if glyph.bounds is None:
            self.fitScaleMetrics()
            return
        fitHeight = self.height()
        fitWidth = self.width()
        left, bottom, right, top = glyph.bounds
        glyphHeight = top - bottom
        glyphHeightPad = glyphHeight + _noPointSizePadding * 2
        glyphWidth = right - left
        glyphWidthPad = glyphWidth + _noPointSizePadding * 2
        self.setScale(min(fitHeight / glyphHeightPad, fitWidth / glyphWidthPad))
        otherWidth = 0
        for glyph_ in self.glyphs():
            if glyph_ == glyph:
                break
            otherWidth += glyph_.width
        dx = (
            0.5 * (fitWidth - glyphWidth * self._scale)
            - (otherWidth + left) * self._scale
        )
        dy = 0.5 * (fitHeight - glyphHeight * self._scale) + top * self._scale
        # TODO: round?
        self._drawingOffset = QPoint(dx, dy)
        self.pointSizeModified.emit(self._impliedPointSize)

    def scrollBy(self, point):
        self._drawingOffset += point
        self.update()

    def verticalBounds(self):
        bottom = self._descender
        top = max(self._capHeight, self._ascender, self._unitsPerEm + self._descender)
        return bottom, top

    def zoom(self, newScale, anchor="center"):
        """
        Zooms the view by *step* increments (with a scale factor of
        1.2^*step*), anchored to *anchor*:

        - QPoint_: center on that point
        - "cursor": center on the mouse cursor position
        - "center": center on the viewport
        - None: don’t anchor, i.e. stick to the viewport’s top-left.

        # TODO: improve docs from QGraphicsView descriptions.

        The default is "center".

        .. _QPoint: http://doc.qt.io/qt-5/qpoint.html
        """
        oldScale = self._scale
        if newScale < 1e-2 or newScale > 1e3:
            return
        # compute new position
        # http://stackoverflow.com/a/32269574/2037879
        if isinstance(anchor, QPoint):
            pos = anchor
        elif anchor == "cursor":
            pos = self.mapFromGlobal(QCursor.pos())
        elif anchor == "center":
            pos = QPoint(0.5 * self.width(), 0.5 * self.height())
        else:
            raise ValueError(f"invalid anchor value: {anchor}")
        deltaToPos = pos / oldScale - self._drawingOffset / oldScale
        delta = deltaToPos * (newScale - oldScale)
        self.setScale(newScale)
        self._drawingOffset -= delta

    # position mapping

    def mapFromCanvas(self, pos, index=None):
        """
        Maps *pos* from glyph canvas to this widget’s coordinates.

        Note that canvas coordinates are scale-independent while widget
        coordinates are not.
        """
        if index is None:
            index = self._activeIndex
        offset = self._drawingOffset
        x, y = pos.x(), pos.y()
        for glyphIndex, glyphRecord in enumerate(self._glyphRecords):
            if glyphIndex == index:
                x += glyphRecord.xOffset
                y += glyphRecord.yOffset
                break
            x += glyphRecord.xAdvance
            y += glyphRecord.yAdvance
        x = x * self._scale + offset.x()
        y = y * -self._scale + offset.y()
        return pos.__class__(x, y)

    def mapToCanvas(self, pos, index=None):
        """
        Maps *pos* from this widget’s to glyph canvas coordinates.

        Note that canvas coordinates are scale-independent while widget
        coordinates are not.
        """
        if index is None:
            index = self._activeIndex
        offset = self._drawingOffset
        x = (pos.x() - offset.x()) * self._inverseScale
        y = (offset.y() - pos.y()) * self._inverseScale
        for glyphIndex, glyphRecord in enumerate(self._glyphRecords):
            if glyphIndex == index:
                x -= glyphRecord.xOffset
                y -= glyphRecord.yOffset
                break
            x -= glyphRecord.xAdvance
            y -= glyphRecord.yAdvance
        return pos.__class__(x, y)

    def mapRectFromCanvas(self, rect):
        x, y, w, h = rect.getRect()
        origin = self.mapFromCanvas(QPointF(x, y))
        w *= self._scale
        h *= self._scale
        return rect.__class__(origin.x(), origin.y() - h, w, h)

    def mapRectToCanvas(self, rect):
        x, y, w, h = rect.getRect()
        origin = self.mapToCanvas(QPointF(x, y))
        w *= self._inverseScale
        h *= self._inverseScale
        return rect.__class__(origin.x(), origin.y() - h, w, h)

    # ---------------
    # Display Control
    # ---------------

    def drawingAttribute(self, attr, flags):
        if flags.isActiveGlyph != flags.isActiveLayer:
            if not flags.isActiveGlyph and attr == "showGlyphComponentStroke":
                return True
            return attr == "showGlyphStroke"
        elif not flags.isActiveGlyph:
            return False
        return self._defaultDrawingAttributes.get(attr)

    def drawingColor(self, attr, flags):
        if attr == "contourFillColor":
            return Qt.black
        return None

    # defaults

    def defaultDrawingAttribute(self, attr):
        return self._defaultDrawingAttributes.get(attr)

    def setDefaultDrawingAttribute(self, attr, value):
        self._defaultDrawingAttributes[attr] = value
        self.update()

    def showFill(self):
        return self.defaultDrawingAttribute("showGlyphFill")

    def setShowFill(self, value):
        self.setDefaultDrawingAttribute("showGlyphFill", value)

    def showStroke(self):
        return self.defaultDrawingAttribute("showGlyphStroke")

    def setShowStroke(self, value):
        self.setDefaultDrawingAttribute("showGlyphStroke", value)

    def showMetrics(self):
        return self.defaultDrawingAttribute("showGlyphMargins")

    def setShowMetrics(self, value):
        self.setDefaultDrawingAttribute("showGlyphMargins", value)
        self.setDefaultDrawingAttribute("showFontVerticalMetrics", value)

    def showImage(self):
        return self.defaultDrawingAttribute("showGlyphImage")

    def setShowImage(self, value):
        self.setDefaultDrawingAttribute("showGlyphImage", value)

    def showMetricsTitles(self):
        return self.defaultDrawingAttribute("showFontVerticalMetricsTitles")

    def setShowMetricsTitles(self, value):
        self.setDefaultDrawingAttribute("showFontVerticalMetricsTitles", value)

    def showGuidelines(self):
        return self.defaultDrawingAttribute("showFontGuidelines")

    def setShowGuidelines(self, value):
        self.setDefaultDrawingAttribute("showFontGuidelines", value)
        self.setDefaultDrawingAttribute("showGlyphGuidelines", value)

    def showOnCurvePoints(self):
        return self.defaultDrawingAttribute("showGlyphOnCurvePoints")

    def setShowOnCurvePoints(self, value):
        self.setDefaultDrawingAttribute("showGlyphStartPoints", value)
        self.setDefaultDrawingAttribute("showGlyphOnCurvePoints", value)

    def showOffCurvePoints(self):
        return self.defaultDrawingAttribute("showGlyphOffCurvePoints")

    def setShowOffCurvePoints(self, value):
        self.setDefaultDrawingAttribute("showGlyphOffCurvePoints", value)

    def showPointCoordinates(self):
        return self.defaultDrawingAttribute("showGlyphPointCoordinates")

    def setShowPointCoordinates(self, value):
        self.setDefaultDrawingAttribute("showGlyphPointCoordinates", value)

    def showAnchors(self):
        return self.defaultDrawingAttribute("showGlyphAnchors")

    def setShowAnchors(self, value):
        self.setDefaultDrawingAttribute("showGlyphAnchors", value)

    def showBlues(self):
        return self.defaultDrawingAttribute("showFontPostscriptBlues")

    def setShowBlues(self, value):
        self.setDefaultDrawingAttribute("showFontPostscriptBlues", value)

    def showFamilyBlues(self):
        return self.defaultDrawingAttribute("showFontPostscriptFamilyBlues")

    def setShowFamilyBlues(self, value):
        self.setDefaultDrawingAttribute("showFontPostscriptFamilyBlues", value)

    def backgroundColor(self):
        return self._backgroundColor

    def setBackgroundColor(self, color):
        self._backgroundColor = color
        self.update()

    # ---------------
    # Drawing helpers
    # ---------------

    def drawGlyphBackground(self, painter, glyph, flags):
        if flags.isActiveLayer:
            # draw the blues
            if self.drawingAttribute("showFontPostscriptBlues", flags):
                self.drawBlues(painter, glyph, flags)
            if self.drawingAttribute("showFontPostscriptFamilyBlues", flags):
                self.drawFamilyBlues(painter, glyph, flags)
            # draw the metrics
            if self.drawingAttribute("showGlyphMetrics", flags):
                self.drawMetrics(painter, glyph, flags)

    def drawBackground(self, painter, index):
        pass

    def drawGlyphLayer(self, painter, glyph, flags):
        # draw the image
        if self.drawingAttribute("showGlyphImage", flags):
            self.drawImage(painter, glyph, flags)
        # draw the guidelines
        if (
            flags.isActiveLayer
            and self.drawingAttribute("showFontGuidelines", flags)
            or self.drawingAttribute("showGlyphGuidelines", flags)
        ):
            self.drawGuidelines(painter, glyph, flags)
        # draw the glyph
        if (
            self.drawingAttribute("showGlyphOnCurvePoints", flags)
            or self.drawingAttribute("showGlyphFill", flags)
            or self.drawingAttribute("showGlyphComponentFill", flags)
        ):
            self.drawFillAndPoints(painter, glyph, flags)
        if self.drawingAttribute("showGlyphStroke", flags) or self.drawingAttribute(
            "showGlyphComponentStroke", flags
        ):
            self.drawStroke(painter, glyph, flags)
        if self.drawingAttribute("showGlyphAnchors", flags):
            self.drawAnchors(painter, glyph, flags)

    def drawForeground(self, painter, index):
        pass

    # drawing primitives

    def drawBlues(self, painter, glyph, flags):
        drawing.drawFontPostscriptBlues(painter, glyph, self._inverseScale)

    def drawFamilyBlues(self, painter, glyph, flags):
        drawing.drawFontPostscriptFamilyBlues(painter, glyph, self._inverseScale)

    def drawMetrics(self, painter, glyph, flags):
        drawVMetrics = flags.isActiveLayer and self.drawingAttribute(
            "showFontVerticalMetrics", flags
        )
        drawText = (
            self.drawingAttribute("showFontVerticalMetricsTitles", flags)
            and self._impliedPointSize > GlyphViewMinSizeForDetails
        )
        drawing.drawGlyphMetrics(
            painter,
            glyph,
            self._inverseScale,
            drawVMetrics=drawVMetrics,
            drawText=drawText,
        )

    def drawImage(self, painter, glyph, flags):
        drawing.drawGlyphImage(painter, glyph, self._inverseScale)

    def drawGuidelines(self, painter, glyph, flags):
        drawText = self._impliedPointSize > GlyphViewMinSizeForDetails
        viewportRect = self.mapRectToCanvas(self.rect()).getRect()
        if self.drawingAttribute("showFontGuidelines", flags):
            drawing.drawFontGuidelines(
                painter, glyph, self._inverseScale, viewportRect, drawText=drawText
            )
        if self.drawingAttribute("showGlyphGuidelines", flags):
            drawing.drawGlyphGuidelines(
                painter, glyph, self._inverseScale, viewportRect, drawText=drawText
            )

    def drawFillAndPoints(self, painter, glyph, flags):
        drawFill = self.drawingAttribute("showGlyphFill", flags)
        drawComponentFill = self.drawingAttribute("showGlyphComponentFill", flags)
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            drawFill=drawFill,
            drawComponentFill=drawComponentFill,
            drawStroke=False,
        )
        if not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawStartPoints = self.drawingAttribute("showGlyphStartPoints", flags)
        drawOnCurves = self.drawingAttribute("showGlyphOnCurvePoints", flags)
        drawOffCurves = self.drawingAttribute("showGlyphOffCurvePoints", flags)
        drawCoordinates = self.drawingAttribute("showGlyphPointCoordinates", flags)
        drawing.drawGlyphPoints(
            painter,
            glyph,
            self._inverseScale,
            drawOnCurves=drawOnCurves,
            drawOffCurves=drawOffCurves,
            drawStartPoints=drawStartPoints,
            drawCoordinates=drawCoordinates,
            backgroundColor=self._backgroundColor,
        )

    def drawStroke(self, painter, glyph, flags):
        drawStroke = self.drawingAttribute("showGlyphStroke", flags)
        drawComponentStroke = self.drawingAttribute("showGlyphComponentStroke", flags)
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            drawFill=False,
            drawComponentsFill=False,
            drawStroke=drawStroke,
            drawComponentStroke=drawComponentStroke,
        )

    def drawAnchors(self, painter, glyph, flags):
        if not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawing.drawGlyphAnchors(painter, glyph, self._inverseScale)

    # ---------------
    # QWidget methods
    # ---------------

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setFont(UIFont)
        painter.setRenderHint(QPainter.Antialiasing)

        # draw the background
        self._glyphRecordsRects = {}
        painter.fillRect(event.rect(), self._backgroundColor)
        if not self._glyphRecords:
            return

        # move into the canvas origin
        offset = self._drawingOffset
        painter.translate(offset)
        painter.scale(self._scale, -self._scale)

        left = offset.x()
        top = offset.y() - self._ascender * self._scale
        height = self._unitsPerEm * self._scale
        painter.save()
        for recordIndex, glyphRecord in enumerate(self._glyphRecords):
            active = recordIndex == self._activeIndex
            xO = glyphRecord.xOffset
            yO = glyphRecord.yOffset
            xA = glyphRecord.xAdvance
            yA = glyphRecord.yAdvance
            # store the glyph rect
            top -= yO * self._scale
            glyphHeight = height + yA * self._scale
            glyphLeft = left + xO * self._scale
            glyphWidth = xA * self._scale
            rect = (glyphLeft, top, glyphWidth, glyphHeight)
            self._glyphRecordsRects[rect] = recordIndex
            # handle placement
            if xO or yO:
                painter.translate(xO, yO)
            # draw the background
            painter.save()
            foreGlyph = glyphRecord.glyph
            foreGlyph = foreGlyph.font[foreGlyph.name]
            self.drawGlyphBackground(painter, foreGlyph, GlyphFlags(active))
            self.drawBackground(painter, recordIndex)
            painter.restore()
            # shift for the next glyph
            painter.translate(xA - xO, yA - yO)
            left += glyphWidth
        painter.restore()

        for recordIndex, glyphRecord in enumerate(self._glyphRecords):
            glyph = glyphRecord.glyph
            xO = glyphRecord.xOffset
            yO = glyphRecord.yOffset
            xA = glyphRecord.xAdvance
            yA = glyphRecord.yAdvance
            # gather the crowd
            layerSet = glyph.layerSet
            if layerSet is None:
                layers = [(glyph, GlyphFlags(True))]
            else:
                layers = []
                for layerName in reversed(layerSet.layerOrder):
                    layer = layerSet[layerName]
                    if glyph.name in layer:
                        layerGlyph = layer[glyph.name]
                        layerFlags = GlyphFlags(
                            recordIndex == self._activeIndex, layerGlyph == glyph
                        )
                        layers.append((layerGlyph, layerFlags))
            # handle placement
            if xO or yO:
                painter.translate(xO, yO)
            # draw layers and foreground
            painter.save()
            for layerGlyph, layerFlags in layers:
                self.drawGlyphLayer(painter, layerGlyph, layerFlags)
            self.drawForeground(painter, recordIndex)
            painter.restore()
            # shift for the next glyph
            painter.translate(xA - xO, yA - yO)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if not hasattr(self, "_drawingOffset"):
            return
        delta = 0.5 * (event.oldSize() - event.size())
        offset = self._drawingOffset
        offset.setX(offset.x() - delta.width())
        offset.setY(offset.y() - delta.height())

    def minimumSizeHint(self):
        return QSize(400, 400)

    def sizeHint(self):
        return QSize(900, 900)

    def showEvent(self, event):
        super().showEvent(event)
        if hasattr(self, "_fitViewport"):
            self.fitScaleBBox()
            del self._fitViewport

    def event(self, event):
        if event.type() == QEvent.Gesture:
            # Handle pan gestures
            panGesture = event.gesture(Qt.PanGesture)
            if panGesture:
                self._drawingOffset += panGesture.delta()

            # Handle pinch gestures
            pinchGesture = event.gesture(Qt.PinchGesture)
            if pinchGesture:
                newScale = self._scale * pinchGesture.scaleFactor()
                self.zoom(newScale, "cursor")
                self.pointSizeModified.emit(self._impliedPointSize)

            return True
        return super().event(event)

    def wheelEvent(self, event):
        if event.modifiers() & platformSpecific.scaleModifier():
            step = event.angleDelta().y() / 120.0
            newScale = self._scale * pow(1.2, step)
            self.zoom(newScale, event.pos())
            self.pointSizeModified.emit(self._impliedPointSize)
        else:
            delta = event.pixelDelta()
            if delta.isNull():
                delta = event.angleDelta()
                dx = delta.x() / 120 * QApplication.wheelScrollLines() * 8
                dy = delta.y() / 120 * QApplication.wheelScrollLines() * 8
                delta = QPoint(dx, dy)
            self._drawingOffset += delta
            self.update()
        event.accept()


class GlyphRecord:
    __slots__ = ["glyph", "cluster", "xOffset", "yOffset", "xAdvance", "yAdvance"]

    def __init__(self):
        self.glyph = None
        self.cluster = 0
        self.xOffset = 0
        self.yOffset = 0
        self.xAdvance = 0
        self.yAdvance = 0

    def __repr__(self):
        return "<GlyphRecord {}={}@{},{}+{}>".format(
            self.glyph.name, self.cluster, self.xAdvance, self.xOffset, self.yOffset
        )



import math
import types

import fontTools
from booleanOperations import union
from defcon import (
    Anchor,
    Component,
    Contour,
    Font,
    Glyph,
    Groups,
    Guideline,
    Image,
    Kerning,
    Layer,
    Point,
)
from fontTools.misc.transform import Identity
from PyQt5.QtWidgets import QApplication
from ufo2ft import compileOTF, compileTTF

from trufont.objects import settings
from trufont.objects.undoManager import UndoManager

_shaper = True
try:
    import harfbuzz  # noqa
    from trufont.objects.layoutEngine import LayoutEngine
except ImportError:
    try:
        import compositor  # noqa
        from defcon import LayoutEngine
    except ImportError:
        _shaper = False


class TFont(Font):
    def __init__(self, *args, **kwargs):
        # TODO: maybe subclass all objects into our own for caller stability
        attrs = (
            ("glyphAnchorClass", TAnchor),
            ("glyphComponentClass", TComponent),
            ("glyphClass", TGlyph),
            ("glyphContourClass", TContour),
            ("glyphPointClass", TPoint),
            ("glyphImageClass", TImage),
            ("guidelineClass", TGuideline),
            ("layerClass", TLayer),
            ("groupsClass", TGroups),
            ("kerningClass", TKerning),
        )
        for attr, defaultClass in attrs:
            if attr not in kwargs:
                kwargs[attr] = defaultClass
        super().__init__(*args, **kwargs)
        self._engine = None

    def __repr__(self):
        return "<{} {} {}>".format(
            self.__class__.__name__, self.info.familyName, self.info.styleName
        )

    @property
    def binaryPath(self):
        if hasattr(self, "_binaryPath"):
            return self._binaryPath
        return None

    @property
    def engine(self):
        if _shaper and self._engine is None:
            self._engine = LayoutEngine(self)
        return self._engine

    @classmethod
    def new(cls):
        font = cls()
        font.info.unitsPerEm = 1000
        font.info.ascender = 750
        font.info.capHeight = 700
        font.info.xHeight = 500
        font.info.descender = -250

        defaultGlyphSet = settings.defaultGlyphSet()
        if defaultGlyphSet:
            glyphNames = None
            glyphSets = settings.readGlyphSets()
            if defaultGlyphSet in glyphSets:
                glyphNames = glyphSets[defaultGlyphSet]
            if glyphNames is not None:
                for name in glyphNames:
                    font.get(name, asTemplate=True)
        font.dirty = False

        app = QApplication.instance()
        data = dict(font=font)
        app.postNotification("newFontCreated", data)

        return font

    def get(self, name, **kwargs):
        return self._glyphSet.get(name, **kwargs)

    def extract(self, path):
        import extractor

        fileFormat = extractor.extractFormat(path)
        app = QApplication.instance()
        data = dict(font=self, format=fileFormat)
        app.postNotification("fontWillExtract", data)
        # don't bring on UndoManager just yet
        func = self.newGlyph
        try:
            self.newGlyph = types.MethodType(Font.newGlyph, self)
            extractor.extractUFO(path, self, fileFormat)
            for glyph in self:
                glyph.dirty = False
                glyph.undoManager = UndoManager(glyph)
        finally:
            self.newGlyph = func
        self.dirty = False
        self._binaryPath = path

    def save(
        self,
        path=None,
        formatVersion=None,
        removeUnreferencedImages=False,
        progressBar=None,
    ):
        app = QApplication.instance()
        data = dict(font=self, path=path or self.path)
        app.postNotification("fontWillSave", data)
        super().save(path, formatVersion, removeUnreferencedImages, progressBar)
        app.postNotification("fontSaved", data)

    def export(self, path, format="otf", compression=None, **kwargs):
        if format == "otf":
            func = compileOTF
        elif format == "ttf":
            func = compileTTF
        else:
            raise ValueError("unknown format: %s" % format)
        if compression is not None:
            invalid = compression - {"none", "woff", "woff2"}
            if invalid:
                raise ValueError(f"unknown compression format: {invalid}")
        # info attrs
        missingAttrs = []
        for attr in (
            "familyName",
            "styleName",
            "unitsPerEm",
            "ascender",
            "descender",
            "xHeight",
            "capHeight",
        ):
            if getattr(self.info, attr) is None:
                missingAttrs.append(attr)
        if missingAttrs:
            raise ValueError(
                "font info attributes required for export are "
                "missing: {}".format(" ".join(missingAttrs))
            )
        # go ahead
        app = QApplication.instance()
        data = dict(
            font=self, path=path, format=format, compression=compression, kwargs=kwargs
        )
        app.postNotification("fontWillExport", data)
        otf = func(self, **kwargs)
        if compression is None or "none" in compression:
            otf.save(path)
        if compression is not None:
            for header in compression:
                if header == "none":
                    continue
                otf.flavor = header
                otf.save(f"{path}.{header}")
        app.postNotification("fontExported", data)

    # sort descriptor

    def _get_sortDescriptor(self):
        # TODO: I'd use defcon sortDescriptor but there is no hard
        # standard for glyphList
        value = self.lib.get("com.typesupply.defcon.sortDescriptor", None)
        if value is not None:
            value = list(value)
        return value

    def _set_sortDescriptor(self, value):
        oldValue = self.lib.get("com.typesupply.defcon.sortDescriptor")
        if value is None or len(value) == 0:
            value = None
            if "com.typesupply.defcon.sortDescriptor" in self.lib:
                del self.lib["com.typesupply.defcon.sortDescriptor"]
        else:
            self.lib["com.typesupply.defcon.sortDescriptor"] = value
        self.postNotification(
            "Font.SortDescriptorChanged", data=dict(oldValue=oldValue, newValue=value)
        )

    sortDescriptor = property(
        _get_sortDescriptor, _set_sortDescriptor, doc="The defcon sort descriptor."
    )


class TLayer(Layer):
    def get(
        self,
        name,
        override=False,
        addUnicode=True,
        asTemplate=False,
        markColor=None,
        width=600,
    ):
        if not override:
            if name in self:
                # TODO: return the glyph here (change dependant code)
                return None
        if asTemplate:
            dirty = self.dirty
            self.disableNotifications(notification=self.changeNotificationName)
        glyph = self.newGlyph(name)
        if asTemplate:
            self.dirty = dirty
            self.enableNotifications(notification=self.changeNotificationName)
        if asTemplate:
            glyph.disableNotifications()
        glyph.width = width
        if addUnicode:
            glyph.autoUnicodes()
        glyph.template = asTemplate
        if asTemplate:
            glyph.dirty = False
            glyph.enableNotifications()
        glyph.markColor = markColor

        app = QApplication.instance()
        data = dict(glyph=glyph)
        app.postNotification("newGlyphCreated", data)

        return glyph

    def loadGlyph(self, name):
        glyph = super().loadGlyph(name)
        glyph.undoManager = UndoManager(glyph)
        return glyph

    def newGlyph(self, name):
        glyph = super().newGlyph(name)
        glyph.undoManager = UndoManager(glyph)
        return glyph

    def _glyphsReloadFilter(self, glyphNames):
        for glyphName in glyphNames:
            if glyphName in self and self[glyphName].template:
                continue
            if glyphName not in self._glyphSet:
                # TODO: blindy treating every KeyError as a "former"
                # template glyph isn't particularly elegant.
                glyph = self[glyphName]
                glyph.clear()
                glyph.template = True
                glyph.dirty = False
                continue
            yield glyphName

    def reloadGlyphs(self, glyphNames):
        super().reloadGlyphs(self._glyphsReloadFilter(glyphNames))

    def saveGlyph(self, glyph, glyphSet, saveAs=False):
        if not glyph.template:
            super().saveGlyph(glyph, glyphSet, saveAs)


class TGlyph(Glyph):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._template = False

    def __repr__(self):
        return "<{} {} ({})>".format(
            self.__class__.__name__, self.name, self.layer.name
        )

    def beginUndoGroup(self, text=None):
        self._undoManager.beginUndoGroup(text)

    def endUndoGroup(self):
        self._undoManager.endUndoGroup()

    # observe anchor selection

    def beginSelfAnchorNotificationObservation(self, anchor):
        if anchor.dispatcher is None:
            return
        super().beginSelfAnchorNotificationObservation(anchor)
        anchor.addObserver(
            observer=self,
            methodName="_selectionChanged",
            notification="Anchor.SelectionChanged",
        )

    def endSelfAnchorNotificationObservation(self, anchor):
        if anchor.dispatcher is None:
            return
        anchor.removeObserver(observer=self, notification="Anchor.SelectionChanged")
        super().endSelfAnchorNotificationObservation(anchor)

    # observe component selection

    def beginSelfComponentNotificationObservation(self, component):
        if component.dispatcher is None:
            return
        super().beginSelfComponentNotificationObservation(component)
        component.addObserver(
            observer=self,
            methodName="_selectionChanged",
            notification="Component.SelectionChanged",
        )

    def endSelfComponentNotificationObservation(self, component):
        if component.dispatcher is None:
            return
        component.removeObserver(
            observer=self, notification="Component.SelectionChanged"
        )
        super().endSelfComponentNotificationObservation(component)

    # observe contours selection

    def selfNotificationCallback(self, notification):
        super().selfNotificationCallback(notification)
        if notification.name == "Glyph.Changed" and self.dirty:
            self.template = False

    def beginSelfContourNotificationObservation(self, contour):
        if contour.dispatcher is None:
            return
        super().beginSelfContourNotificationObservation(contour)
        contour.addObserver(
            observer=self,
            methodName="_selectionChanged",
            notification="Contour.SelectionChanged",
        )

    def endSelfContourNotificationObservation(self, contour):
        if contour.dispatcher is None:
            return
        contour.removeObserver(observer=self, notification="Contour.SelectionChanged")
        super().endSelfContourNotificationObservation(contour)

    def _selectionChanged(self, notification):
        if self.dispatcher is None:
            return
        self.postNotification(notification="Glyph.SelectionChanged")

    def _get_selected(self):
        for contour in self:
            if not contour.selected:
                return False
        return True

    def _set_selected(self, value):
        for contour in self:
            contour.selected = value

    selected = property(
        _get_selected,
        _set_selected,
        doc="The selected state of the contour. "
        "Selected state corresponds to all children points being selected."
        "Set selected state to select or unselect all points in the glyph.",
    )

    def _get_selection(self):
        selection = set()
        for contour in self:
            selection.update(contour.selection)
        return selection

    def _set_selection(self, selection):
        if selection == self.selection:
            return
        for contour in self:
            contour.selection = selection

    selection = property(
        _get_selection,
        _set_selection,
        doc="A list of children points that are selected.",
    )

    def _get_template(self):
        return self._template

    def _set_template(self, value):
        self._template = value

    template = property(
        _get_template,
        _set_template,
        doc="A boolean indicating whether the glyph is a template glyph.",
    )

    def _get_side1KerningGroup(self):
        font = self.font
        if font is None:
            return
        groups = font.groups
        return groups.side1GroupForGlyphName(self.name)

    side1KerningGroup = property(
        _get_side1KerningGroup, doc="The left side kerning group of the glyph."
    )

    def _get_side2KerningGroup(self):
        font = self.font
        if font is None:
            return
        groups = font.groups
        return groups.side2GroupForGlyphName(self.name)

    side2KerningGroup = property(
        _get_side2KerningGroup, doc="The right side kerning group of the glyph."
    )

    def autoUnicodes(self):
        app = QApplication.instance()
        if app.GL2UV is not None:
            GL2UV = app.GL2UV
        else:
            GL2UV = fontTools.agl.AGL2UV
        hexes = "ABCDEF0123456789"
        name = self.name
        if name in GL2UV:
            uni = GL2UV[name]
        elif (
            name.startswith("uni")
            and len(name) == 7
            and all(c in hexes for c in name[3:])
        ):
            uni = int(name[3:], 16)
        elif (
            name.startswith("u")
            and len(name) in (5, 7)
            and all(c in hexes for c in name[1:])
        ):
            uni = int(name[1:], 16)
        else:
            return
        self.unicodes = [uni]

    def rename(self, newName):
        if self.layerSet is not None:
            font = self.font
            oldName = self.name
            font.disableNotifications()
            for layer in self.layerSet:
                if oldName in layer:
                    glyph = layer[oldName]
                    glyph.name = newName
            font.enableNotifications()
        else:
            self.name = newName

    def hasOverlap(self):
        closed = []
        length = len(self)
        for contour in self:
            if contour.open:
                length -= 1
            else:
                closed.append(contour)
        glyph = self.__class__()
        union(closed, glyph.getPointPen())
        return len(glyph) != length

    def removeOverlap(self):
        open_, closed = [], []
        for contour in self:
            (open_ if contour.open else closed).append(contour)
        self.beginUndoGroup()
        self.clearContours()
        pointPen = self.getPointPen()
        union(closed, pointPen)
        for contour in open_:
            contour.drawPoints(pointPen)
        self.endUndoGroup()

    def scale(self, pt, center=(0, 0)):
        dx, dy = center
        x, y = pt
        sT = Identity.translate(dx, dy)
        sT = sT.scale(x=x, y=y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def transform(self, matrix):
        self.beginUndoGroup()
        for contour in self:
            contour.transform(matrix)
        for component in self.components:
            component.transform(matrix)
        for anchor in self.anchors:
            anchor.transform(matrix)
        for guideline in self.guidelines:
            guideline.transform(matrix)
        self.endUndoGroup()

    def rotate(self, angle, offset=(0, 0)):
        dx, dy = offset
        radAngle = math.radians(angle)
        rT = Identity.translate(dx, dy)
        rT = rT.rotate(radAngle)
        rT = rT.translate(-dx, -dy)
        self.transform(rT)

    def skew(self, angle, offset=(0, 0)):
        dx, dy = offset
        x, y = angle
        x, y = math.radians(x), math.radians(y)
        sT = Identity.translate(dx, dy)
        sT = sT.skew(x, y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def snap(self, base):
        self.beginUndoGroup()
        for contour in self:
            contour.snap(base)
        for component in self.components:
            component.snap(base)
        for anchor in self.anchors:
            anchor.snap(base)
        for guideline in self.guidelines:
            guideline.snap(base)
        self.endUndoGroup()


class TKerning(Kerning):
    def find(self, firstGlyph, secondGlyph):
        first = firstGlyph.name
        second = secondGlyph.name
        firstGroup = firstGlyph.side2KerningGroup
        secondGroup = secondGlyph.side1KerningGroup
        # make an ordered list of pairs to look up
        pairs = [
            (first, second),
            (first, secondGroup),
            (firstGroup, second),
            (firstGroup, secondGroup),
        ]
        # look up the pairs and return any matches
        for pair in pairs:
            if pair in self:
                return self[pair]
        return 0

    def write(self, firstGlyph, secondGlyph, value):
        first = firstGlyph.name
        second = secondGlyph.name
        firstGroup = firstGlyph.side2KerningGroup
        secondGroup = secondGlyph.side1KerningGroup
        pairs = [
            (first, second),
            (first, secondGroup),
            (firstGroup, second),
            (firstGroup, secondGroup),
        ]
        for pair in pairs:
            if pair in self:
                self[pair] = value
                return
        self[pairs[3 if firstGroup and secondGroup else 0]] = value


class TGroups(Groups):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _bootstrapGroupsCache(self):
        self._buildGroupsCache()
        # TODO: removeObserver
        self.addObserver(self, "_groupSet", "Groups.GroupSet")
        self.addObserver(self, "_groupDeleted", "Groups.GroupDeleted")
        self.addObserver(self, "_buildGroupsCache", "Groups.Cleared")
        self.addObserver(self, "_buildGroupsCache", "Groups.Updated")

    def _buildGroupsCache(self, *_):
        self._side1Groups = dict()
        self._side2Groups = dict()
        for group in self.keys():
            if group.startswith("public.kern1."):
                stor = self._side1Groups
            elif group.startswith("public.kern2."):
                stor = self._side2Groups
            else:
                continue
            for glyphName in self[group]:
                if glyphName is None:
                    continue
                stor[glyphName] = group

    def _groupSet(self, notification):
        group = notification.data["key"]
        if group.startswith("public.kern1."):
            stor = self._side1Groups
        elif group.startswith("public.kern2"):
            stor = self._side2Groups
        else:
            return
        oldValue = set(notification.data["oldValue"])
        value = set(notification.data["value"])
        for delName in oldValue - value:
            del stor[delName]
        for addName in value - oldValue:
            if addName is None:
                continue
            stor[addName] = group

    def _groupDeleted(self, notification):
        group = notification.data["key"]
        if not group.startswith("public.kern"):
            return
        for stor in (self._side1Groups, self._side2Groups):
            for glyphName, group_ in stor.items():
                if group_ == group:
                    del stor[glyphName]

    def side1GroupForGlyphName(self, glyphName):
        if not hasattr(self, "_side1Groups"):
            self._bootstrapGroupsCache()
        return self._side1Groups.get(glyphName)

    def side2GroupForGlyphName(self, glyphName):
        if not hasattr(self, "_side2Groups"):
            self._bootstrapGroupsCache()
        return self._side2Groups.get(glyphName)


class TContour(Contour):
    def __init__(self, *args, **kwargs):
        if "pointClass" not in kwargs:
            kwargs["pointClass"] = TPoint
        super().__init__(*args, **kwargs)

    def _get_selected(self):
        for point in self:
            if not point.selected:
                return False
        return True

    def _set_selected(self, value):
        for point in self:
            point.selected = value
        self.postNotification(notification="Contour.SelectionChanged")

    selected = property(
        _get_selected,
        _set_selected,
        doc="The selected state of the contour. "
        "Selected state corresponds to all children points being selected.",
    )

    def _get_selection(self):
        selection = set()
        for point in self:
            if point.selected:
                selection.add(point)
        return selection

    def _set_selection(self, selection):
        if selection == self.selection:
            return
        for point in self:
            point.selected = point in selection
        self.postNotification(notification="Contour.SelectionChanged")

    selection = property(
        _get_selection,
        _set_selection,
        doc="A list of children points that are selected.",
    )

    def drawPoints(self, pointPen):
        """
        Draw the contour with **pointPen**.
        """
        pointPen.beginPath()
        for point in self._points:
            pointPen.addPoint(
                (point.x, point.y),
                segmentType=point.segmentType,
                smooth=point.smooth,
                name=point.name,
                selected=point.selected,
            )
        pointPen.endPath()

    def getPoint(self, index):
        return self[index % len(self)]

    def scale(self, pt, center=(0, 0)):
        dx, dy = center
        x, y = pt
        sT = Identity.translate(dx, dy)
        sT = sT.scale(x, y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def transform(self, matrix):
        for point in self:
            point.x, point.y = matrix.transformPoint((point.x, point.y))
        self.dirty = True

    def snap(self, base):
        for point in self:
            point.x = _snap(point.x, base)
            point.y = _snap(point.y, base)
        self.dirty = True


class TAnchor(Anchor):
    def __init__(self, *args, **kwargs):
        self._selected = False
        super().__init__(*args, **kwargs)
        if "anchorDict" in kwargs:
            anchorDict = kwargs["anchorDict"]
            if anchorDict is not None:
                self._selected = anchorDict.get("selected", False)

    def _get_selected(self):
        return self._selected

    def _set_selected(self, value):
        if value == self._selected:
            return
        self._selected = value
        self.postNotification(notification="Anchor.SelectionChanged")

    # TODO: add to repr
    selected = property(
        _get_selected,
        _set_selected,
        doc="A boolean indicating the selected state of the anchor.",
    )

    def scale(self, pt, center=(0, 0)):
        dx, dy = center
        x, y = pt
        sT = Identity.translate(dx, dy)
        sT = sT.scale(x, y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def transform(self, matrix):
        self.x, self.y = matrix.transformPoint((self.x, self.y))

    def snap(self, base):
        self.x = _snap(self.x, base)
        self.y = _snap(self.y, base)


class TComponent(Component):
    def __init__(self, *args, **kwargs):
        self._selected = False
        super().__init__(*args, **kwargs)

    def _get_selected(self):
        return self._selected

    def _set_selected(self, value):
        if value == self._selected:
            return
        self._selected = value
        self.postNotification(notification="Component.SelectionChanged")

    selected = property(
        _get_selected,
        _set_selected,
        doc="A boolean indicating the selected state of the component.",
    )

    def scale(self, pt, center=(0, 0)):
        dx, dy = center
        x, y = pt
        sT = Identity.translate(dx, dy)
        sT = sT.scale(x, y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def transform(self, matrix):
        self.transformation = tuple(matrix.transform(self.transformation))

    def snap(self, base):
        xScale, xyScale, yxScale, yScale, xOffset, yOffset = self._transformation
        xOffset = _snap(xOffset, base)
        yOffset = _snap(yOffset, base)
        # TODO: should scale be snapped too?
        self.transformation = (xScale, xyScale, yxScale, yScale, xOffset, yOffset)


class TPoint(Point):
    __slots__ = ["_selected"]

    def __init__(self, pt, selected=False, **kwargs):
        super().__init__(pt, **kwargs)
        self._selected = selected

    def _get_selected(self):
        return self._selected

    def _set_selected(self, value):
        self._selected = value

    selected = property(
        _get_selected,
        _set_selected,
        doc="A boolean indicating the selected state of the point.",
    )


class TGuideline(Guideline):
    def __init__(self, *args, **kwargs):
        self._selected = False
        super().__init__(*args, **kwargs)
        if "guidelineDict" in kwargs:
            guidelineDict = kwargs["guidelineDict"]
            if guidelineDict is not None:
                self._selected = guidelineDict.get("selected", False)

    def _get_selected(self):
        return self._selected

    def _set_selected(self, value):
        if value == self._selected:
            return
        self._selected = value
        self.postNotification(notification="Guideline.SelectionChanged")

    # TODO: add to repr
    selected = property(
        _get_selected,
        _set_selected,
        doc="A boolean indicating the selected state of the guideline.",
    )

    def scale(self, pt, center=(0, 0)):
        dx, dy = center
        x, y = pt
        sT = Identity.translate(dx, dy)
        sT = sT.scale(x, y)
        sT = sT.translate(-dx, -dy)
        self.transform(sT)

    def transform(self, matrix):
        self.x, self.y = matrix.transformPoint((self.x, self.y))

    def snap(self, base):
        self.x = _snap(self.x, base)
        self.y = _snap(self.y, base)


class TImage(Image):
    def __init__(self, *args, **kwargs):
        self._selected = False
        super().__init__(*args, **kwargs)
        if "imageDict" in kwargs:
            imageDict = kwargs["imageDict"]
            if imageDict is not None:
                self._selected = imageDict.get("selected", False)

    def _get_selected(self):
        return self._selected

    def _set_selected(self, value):
        if value == self._selected:
            return
        self._selected = value
        self.postNotification(notification="Image.SelectionChanged")

    selected = property(
        _get_selected,
        _set_selected,
        doc="A boolean indicating the selected state of the anchor.",
    )


def _snap(x, base=5):
    if not base:
        return x
    return base * round(x / base)



from PyQt5.QtCore import QAbstractItemModel, QModelIndex, QSize, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QTreeView, QVBoxLayout, QWidget

from trufont.objects import settings


class KerningDictModel(QAbstractItemModel):
    def __init__(self, mapping={}, parent=None):
        super().__init__(parent)
        self.setupModelData(mapping)

    def setupModelData(self, mapping):
        self.layoutAboutToBeChanged.emit()
        self._data = kerns = dict()
        for key, value in mapping.items():
            kern1, kern2 = key
            kern1 = kern1.replace("public.kern1.", "@")
            kern2 = kern2.replace("public.kern2.", "@")
            ct = (kern2, value)
            if kern1 in kerns:
                kerns[kern1].append(ct)
            else:
                kerns[kern1] = [ct]
        self._keys = list(self._data.keys())
        self.layoutChanged.emit()

    def columnCount(self, parent):
        return 2

    def data(self, index, role):
        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            key = index.internalPointer()
            if key is None:
                if index.column() > 0:
                    return None
                return self._keys[index.row()]
            return self._data[key][index.row()][index.column()]
        elif role == Qt.ForegroundRole:
            data = index.data()
            if isinstance(data, str) and data.startswith("@"):
                return QColor(5, 5, 96)

        return None

    def index(self, row, column, parent):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if parent.isValid():
            key = self._keys[parent.row()]
            return self.createIndex(row, column, key)
        return self.createIndex(row, column, None)

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        key = index.internalPointer()
        if key is None:
            return QModelIndex()
        row = self._keys.index(key)
        return self.createIndex(row, 0, None)

    def rowCount(self, parent=QModelIndex()):
        if not parent.isValid():
            return len(self._data)
        key = parent.internalPointer()
        if key is not None:
            # child item
            return 0
        key = self._keys[parent.row()]
        return len(self._data[key])


class KerningWindow(QWidget):
    def __init__(self, font, parent=None):
        super().__init__(parent, Qt.Window)
        self._font = font
        self._font.kerning.addObserver(self, "_kerningChanged", "Kerning.Changed")
        self._font.info.addObserver(self, "_fontInfoChanged", "Info.Changed")
        self.kerningView = QTreeView(self)
        self.kerningView.setModel(KerningDictModel(font.kerning, self.kerningView))
        self.kerningView.expandAll()
        metrics = self.kerningView.fontMetrics()
        self.kerningView.setColumnWidth(1, 8 * metrics.width("0"))
        hdr = self.kerningView.header()
        hdr.setStretchLastSection(False)
        hdr.setSectionResizeMode(0, hdr.Stretch)
        hdr.hide()

        layout = QVBoxLayout(self)
        layout.addWidget(self.kerningView)
        layout.setContentsMargins(0, 0, 0, 0)

        self.updateWindowTitle(font=font)
        self.readSettings()

    def readSettings(self):
        geometry = settings.kerningWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setKerningWindowGeometry(self.saveGeometry())

    def updateWindowTitle(self, title=None, font=None):
        if title is None:
            title = self.tr("Kerning")
        if font is not None:
            title = "{} – {} {}".format(
                title, font.info.familyName, font.info.styleName
            )
        self.setWindowTitle(title)

    # -------------
    # Notifications
    # -------------

    def _kerningChanged(self, notification):
        model = self.kerningView.model()
        model.setupModelData(self._font.kerning)

    def _fontInfoChanged(self, notification):
        self.updateWindowTitle(font=self._font)

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(280, 460)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self._font.kerning.removeObserver(self, "Kerning.Changed")
            self._font.info.removeObserver(self, "Info.Changed")



import functools
import pickle
import weakref

from PyQt5.QtCore import QCoreApplication, QObject, pyqtSignal

tr = functools.partial(QCoreApplication.translate, "UndoManager")
_valueNotifications = n = dict()
n["Glyph.NameChanged"] = tr("Name changed.")
n["Glyph.UnicodesChanged"] = tr("Unicode(s) changed.")
n["Glyph.WidthChanged"] = tr("Width changed.")
n["Glyph.HeightChanged"] = tr("Height changed.")
n["Glyph.NoteChanged"] = tr("Note changed.")
_contentNotifications = n = dict()
n["Glyph.ContoursChanged"] = tr("Contours changed.")
n["Glyph.ComponentsChanged"] = tr("Components changed.")
n["Glyph.AnchorsChanged"] = tr("Anchors changed.")
n["Glyph.GuidelinesChanged"] = tr("Guidelines changed.")
n["Glyph.ImageChanged"] = tr("Image changed.")
del n
del tr


def _attrForNotification(name):
    attr = name[6:-7].lower()
    if attr == "contours":
        attr = "_" + attr
    return attr


def _setGlyphContent(glyph, attr, value):
    data = pickle.loads(value)
    glyph.holdNotifications()
    if attr == "_contours":
        glyph.clearContours()
        shallow = data and "pen" not in data[0]  # XXX: flimsy
        if shallow:
            glyph._shallowLoadedContours = data
        else:
            for element in data:
                contour = glyph.instantiateContour()
                contour.setDataFromSerialization(element)
                glyph.appendContour(contour)
    elif attr == "components":
        glyph.clearComponents()
        for element in data:
            component = glyph.instantiateComponent()
            component.setDataFromSerialization(element)
            glyph.appendComponent(component)
    elif attr == "anchors":
        glyph.clearAnchors()
        for element in data:
            anchor = glyph.instantiateAnchor()
            anchor.setDataFromSerialization(element)
            glyph.appendAnchor(anchor)
    elif attr == "guidelines":
        glyph.clearGuidelines()
        for element in data:
            guideline = glyph.instantiateGuideline()
            guideline.setDataFromSerialization(element)
            glyph.appendGuideline(guideline)
    elif attr == "image":
        image = glyph.instantiateImage()
        image.setDataFromSerialization(data)
        glyph.image = image
    glyph.releaseHeldNotifications()


# TODO: limit the number of elements


class UndoManager(QObject):
    canUndoChanged = pyqtSignal(bool)
    canRedoChanged = pyqtSignal(bool)
    # undoTextChanged = pyqtSignal(str)
    # redoTextChanged = pyqtSignal(str)

    def __init__(self, glyph):
        super().__init__()
        self._glyph = weakref.ref(glyph)
        self._init()

        self._subscribeToGlyph()

    def _init(self):
        self._undoStack = []
        self._redoStack = []
        self._cleanIndex = 0
        self._dumps = dict()
        self._stashedNotifications = dict()
        self._undoGroups = 0

    def _subscribeToGlyph(self):
        glyph = self.glyph
        for name in _valueNotifications.keys():
            glyph.addObserver(self, "_valueChanged", name)
        for name in _contentNotifications.keys():
            attr = _attrForNotification(name)
            data = None
            if attr == "_contours":
                data = getattr(glyph, "_shallowLoadedContours", None)
            elif attr == "image":
                data = getattr(glyph, attr).getDataForSerialization()
            if data is None:
                data = [item.getDataForSerialization() for item in getattr(glyph, attr)]
            self._dumps[name] = pickle.dumps(data)
            glyph.addObserver(self, "_contentChanged", name)

    def _unsubscribeFromGlyph(self):
        glyph = self.glyph
        for name in _valueNotifications.keys():
            glyph.removeObserver(self, name)
        for name in _contentNotifications.keys():
            glyph.removeObserver(self, name)
        self._dumps = dict()

    @property
    def glyph(self):
        if self._glyph is None:
            return None
        return self._glyph()

    # -------------
    # Notifications
    # -------------

    def _valueChanged(self, notification):
        name = notification.name
        data = notification.data
        if self._undoGroups:
            if name not in self._stashedNotifications:
                self._stashedNotifications[name] = data
            else:
                self._stashedNotifications[name]["newValue"] = data["newValue"]
        else:
            self._pushValueChange(name, data)

    def _contentChanged(self, notification):
        name = notification.name
        if self._undoGroups:
            self._stashedNotifications[name] = None
        else:
            self._pushContentChange(name)

    # handlers

    def _pushValueChange(self, name, data):
        undoWasLocked = not self.canUndo()
        redoWasEnabled = self.canRedo()

        self._redoStack = []
        self._undoStack.append((name, data))
        if undoWasLocked:
            self.canUndoChanged.emit(True)
        if redoWasEnabled:
            self.canRedoChanged.emit(False)

    def _pushContentChange(self, name):
        glyph = self.glyph
        attr = _attrForNotification(name)
        if name not in self._dumps:
            oldValue = self._dumps["_shallowLoadedContours"]
            del self._dumps["_shallowLoadedContours"]
        else:
            oldValue = self._dumps[name]
        if attr == "image":
            data = getattr(glyph, attr).getDataForSerialization()
        else:
            data = [item.getDataForSerialization() for item in getattr(glyph, attr)]
        newValue = self._dumps[name] = pickle.dumps(data)
        data = dict(oldValue=oldValue, newValue=newValue)

        self._pushValueChange(name, data)

    # ----------
    # Public API
    # ----------

    def clear(self):
        self._init()

    # basic API

    def beginUndoGroup(self, text=None):
        if not self._undoGroups:
            self._undoGroupText = text
        self._undoGroups += 1

    def endUndoGroup(self):
        if not self._undoGroups:
            print("warning: unmatched endUndoGroup()")
            return
        self._undoGroups -= 1
        if not self._undoGroups:
            if self._stashedNotifications:
                undoStack = self._undoStack
                self._undoStack = group = []
                for name, data in self._stashedNotifications.items():
                    if data is None:
                        self._pushContentChange(name)
                    else:
                        self._pushValueChange(name, data)
                self._stashedNotifications = dict()
                self._undoStack = undoStack
                self._undoStack.append((self._undoGroupText, group))
            del self._undoGroupText

    def canUndo(self):
        return not self._undoGroups and bool(self._undoStack)

    def canRedo(self):
        return not self._undoGroups and bool(self._redoStack)

    def undo(self):
        if not self._undoStack or self._undoGroups:
            return
        glyph = self.glyph
        redoWasLocked = not self.canRedo()

        # pop the undo element
        content = self._undoStack.pop()
        if isinstance(content[1], list):
            elements = content[1]
        else:
            elements = (content,)
        for element in reversed(elements):
            # apply the old value
            name, data = element
            attr = _attrForNotification(name)
            value = data["oldValue"]
            glyph.disableNotifications(observer=self)
            if name in _contentNotifications:
                _setGlyphContent(glyph, attr, value)
                self._dumps[name] = value
            else:
                setattr(glyph, attr, value)
            glyph.enableNotifications(observer=self)
        # push as redo element
        self._redoStack.append(content)

        if len(self._undoStack) == self._cleanIndex:
            glyph.dirty = False
        if redoWasLocked:
            self.canRedoChanged.emit(True)
        if not self.canUndo():
            self.canUndoChanged.emit(False)

    def redo(self):
        if not self._redoStack or self._undoGroups:
            return
        glyph = self.glyph
        undoWasLocked = not self.canUndo()

        # pop the redo element
        content = self._redoStack.pop()
        if isinstance(content[1], list):
            elements = content[1]
        else:
            elements = (content,)
        for element in elements:
            # apply the new value
            name, data = element
            attr = _attrForNotification(name)
            value = data["newValue"]
            glyph.disableNotifications(observer=self)
            if name in _contentNotifications:
                _setGlyphContent(glyph, attr, value)
                self._dumps[name] = value
            else:
                setattr(glyph, attr, value)
            glyph.enableNotifications(observer=self)
        # push as undo element
        self._undoStack.append(content)

        if len(self._undoStack) == self._cleanIndex:
            glyph.dirty = False
        if undoWasLocked:
            self.canUndoChanged.emit(True)
        if not self.canRedo():
            self.canRedoChanged.emit(False)

    # clean

    def isClean(self):
        return self._cleanIndex == len(self._undoStack)

    def setClean(self):
        self._cleanIndex = len(self._undoStack)

    # text

    def undoText(self):
        name = self._undoStack[-1][0]
        if name in _contentNotifications:
            return _contentNotifications[name]
        return _valueNotifications[name]

    def redoText(self):
        name = self._redoStack[-1][0]
        if name in _contentNotifications:
            return _contentNotifications[name]
        return _valueNotifications[name]
